import sys
from importlib import import_module
from types import ModuleType
from ..errors import get_exception_text
from . import __name__ as _module_name


def _print_help() -> None:
    print(
        "Usage:\n"
        "  daves-dev-tools git <command> [options]\n\n"
        "Commands:\n"
        "  download                    Download files from a GIT repo "
        "matching one or more\n"
        "                              specified pattern(s).\n"
        "  tag-version                 Tag your repo with the python package "
        "version, if a tag for\n"
        "                              that version doesn't already exist."
    )


def _get_command() -> str:
    command: str = ""
    if len(sys.argv) > 1:
        command = sys.argv.pop(1).lower().replace("-", "_")
    return command


def main() -> None:
    """
    Run a sub-module `main` function.
    """
    command = _get_command()
    module: ModuleType
    try:
        try:
            module = import_module(f"{_module_name}.{command}.__main__")
        except ImportError:
            module = import_module(f"{_module_name}.{command}")
        module.main()  # type: ignore
    except Exception:
        print(get_exception_text())
        _print_help()


if __name__ == "__main__":
    main()



import os
import argparse
import tomli
import tomli_w
import re
from io import StringIO
from dataclasses import dataclass
from pkg_resources import Distribution
from configparser import ConfigParser, SectionProxy
from typing import (
    Dict,
    Iterable,
    IO,
    List,
    Callable,
    Set,
    Union,
    Tuple,
    Any,
    Optional,
)
from packaging.specifiers import Specifier, SpecifierSet
from packaging.requirements import Requirement
from packaging.version import Version, LegacyVersion, parse as parse_version
from more_itertools import unique_everseen
from ..utilities import iter_parse_delimited_values
from .utilities import (
    normalize_name,
    get_installed_distributions,
    is_requirement_string,
)


def get_updated_requirement_string(
    requirement_string: str, ignore: Iterable[str] = ()
) -> str:
    """
    This function accepts a requirement string, and returns a requirement
    string updated to reflect the version of the requirement installed
    in the current environment
    """
    return _get_updated_requirement_string(
        requirement_string, set(map(normalize_name, ignore))
    )


@dataclass
class _Version:
    """
    Instances of this class can be be passed as `self` in a call
    to `packaging.version.Version.__str__`, and thereby can facilitate
    operations to mimic mutability for the aforementioned class.
    """

    epoch: int
    release: Tuple[int, ...]
    pre: Any
    post: Any
    dev: Any
    local: Any


def _update_requirement_specifiers(
    requirement: Requirement, installed_version_string: str
) -> None:
    """
    This function updates specifier version numbers for a requirement
    to match the installed version of the package
    """
    installed_version: Union[Version, LegacyVersion] = parse_version(
        installed_version_string
    )
    specifier: Specifier
    updated_specifier_strings: List[str] = []
    for specifier in requirement.specifier:  # type: ignore
        # Only update requirement to match our installed version
        # if the requirement is *inclusive*
        if ("=" in specifier.operator) and ("!" not in specifier.operator):
            specifier_version: Union[Version, LegacyVersion] = parse_version(
                specifier.version
            )
            assert installed_version.release is not None
            if specifier_version.release is None:
                updated_specifier_strings.append(f"{specifier.operator}")
            else:
                greater_or_equal_specificity: bool = len(
                    specifier_version.release
                ) >= len(installed_version.release)
                specifier_version_data: _Version = _Version(
                    epoch=installed_version.epoch,
                    # Truncate the updated version requirement at the same
                    # level of specificity as the old
                    release=installed_version.release[
                        : len(specifier_version.release)
                    ],
                    pre=(
                        installed_version.pre
                        if greater_or_equal_specificity
                        else None
                    ),
                    post=(
                        installed_version.post
                        if greater_or_equal_specificity
                        else None
                    ),
                    dev=(
                        installed_version.dev
                        if greater_or_equal_specificity
                        else None
                    ),
                    local=(
                        installed_version.local
                        if greater_or_equal_specificity
                        else None
                    ),
                )
                version_string: str = Version.__str__(
                    specifier_version_data  # type: ignore
                )
                updated_specifier_strings.append(
                    f"{specifier.operator}{version_string}"
                )
        else:
            updated_specifier_strings.append(str(specifier))
    requirement.specifier = SpecifierSet(",".join(updated_specifier_strings))


def _get_updated_requirement_string(
    requirement_string: str, ignore: Set[str]
) -> str:
    """
    This function updates version numbers in a requirement string to match
    those installed in the current environment
    """
    # Skip empty requirement strings
    if not is_requirement_string(requirement_string):
        return requirement_string
    requirement: Requirement = Requirement(requirement_string)
    name: str = normalize_name(requirement.name)
    if name in ignore:
        return requirement_string
    try:
        distribution: Distribution = get_installed_distributions()[name]
        _update_requirement_specifiers(requirement, distribution.version)
    except KeyError:
        # If the requirement isn't installed, we can't update the version
        pass
    return str(requirement)


def _normalize_ignore_argument(ignore: Iterable[str]) -> Set[str]:
    ignore_set: Set[str]
    # Normalize/harmonize excluded project names
    if isinstance(ignore, str):
        ignore = (ignore,)
    ignore_set = set(map(normalize_name, ignore))
    return ignore_set


def get_updated_requirements_txt(data: str, ignore: Iterable[str] = ()) -> str:
    """
    Return the contents of a **requirements.txt** file, updated to reflect the
    currently installed project versions, excluding those specified in
    `ignore`.

    Parameters:

    - data (str): The contents of a **setup.cfg** file
    - ignore ([str]): One or more project names to leave as-is
    """
    ignore_set: Set[str] = _normalize_ignore_argument(ignore)

    def get_updated_requirement_string(requirement: str) -> str:
        return _get_updated_requirement_string(requirement, ignore=ignore_set)

    return "\n".join(map(get_updated_requirement_string, data.split("\n")))


def get_updated_setup_cfg(
    data: str, ignore: Iterable[str] = (), all_extra_name: str = ""
) -> str:
    """
    Return the contents of a **setup.cfg** file, updated to reflect the
    currently installed project versions, excluding those specified in
    `ignore`.

    Parameters:

    - data (str): The contents of a **setup.cfg** file
    - ignore ([str]): One or more project names to leave as-is
    - all_extra_name (str): A name under which to store an extra which
      consolidates requirements from all other extras
    """
    ignore_set: Set[str] = _normalize_ignore_argument(ignore)

    def get_updated_requirement_string(requirement: str) -> str:
        return _get_updated_requirement_string(requirement, ignore=ignore_set)

    # Parse
    parser: ConfigParser = ConfigParser()
    parser.read_string(data)
    # Update
    if ("options" in parser) and ("install_requires" in parser["options"]):
        parser["options"]["install_requires"] = "\n".join(
            map(  # type: ignore
                get_updated_requirement_string,
                parser["options"]["install_requires"].split("\n"),
            )
        )
    if "options.extras_require" in parser:
        extras_require: SectionProxy = parser["options.extras_require"]
        all_extra_requirements: List[str] = []
        extra_name: str
        extra_requirements_string: str
        extra_requirements: List[str]
        for extra_name, extra_requirements_string in extras_require.items():
            if extra_name != all_extra_name:
                extra_requirements = list(
                    map(
                        get_updated_requirement_string,
                        extra_requirements_string.split("\n"),
                    )
                )
                if all_extra_name:
                    all_extra_requirements += extra_requirements
                extras_require[extra_name] = "\n".join(extra_requirements)
        # If a name was specified for an all-encompasing extra,
        # we de-duplicate and update or create that extra
        if all_extra_name:
            # We pre-pend an empty requirement string in order to]
            # force new-line creation at the beginning of the extra
            extras_require[all_extra_name] = "\n".join(
                unique_everseen([""] + all_extra_requirements)
            )
    # Return as a string
    setup_cfg: str
    setup_cfg_io: IO[str]
    with StringIO() as setup_cfg_io:
        parser.write(setup_cfg_io)
        setup_cfg_io.seek(0)
        setup_cfg = re.sub(r"[ ]+(\n|$)", r"\1", setup_cfg_io.read()).strip()
        return f"{setup_cfg}\n"


def get_updated_tox_ini(data: str, ignore: Iterable[str] = ()) -> str:
    """
    Return the contents of a **tox.ini** file, updated to reflect the
    currently installed project versions, excluding those specified in
    `ignore`.

    Parameters:

    - data (str): The contents of a **tox.ini** file
    - ignore ([str]): One or more project names to leave as-is
    """
    ignore_set: Set[str] = _normalize_ignore_argument(ignore)

    def get_updated_requirement_string(requirement: str) -> str:
        prefix: Optional[str] = None
        if ":" in requirement:
            prefix, requirement = requirement.split(":", maxsplit=1)
        requirement = _get_updated_requirement_string(
            requirement, ignore=ignore_set
        )
        if prefix is not None:
            requirement = f"{prefix}: {requirement.lstrip()}"
        return requirement

    # Parse
    parser: ConfigParser = ConfigParser()
    parser.read_string(data)

    def update_section_options(section_name: str, option_name: str) -> None:
        if parser.has_option(section_name, option_name):
            parser.set(
                section_name,
                option_name,
                "\n".join(
                    map(
                        get_updated_requirement_string,
                        parser.get(section_name, option_name).split("\n"),
                    )
                ),
            )

    def update_section(section_name: str) -> None:
        update_section_options(section_name, "deps")
        if section_name == "tox":
            update_section_options(section_name, "requires")

    # Update
    list(map(update_section, parser.sections()))
    # Return as a string
    tox_ini: str
    tox_ini_io: IO[str]
    with StringIO() as tox_ini_io:
        parser.write(tox_ini_io)
        tox_ini_io.seek(0)
        tox_ini = re.sub(r"[ ]+(\n|$)", r"\1", tox_ini_io.read()).strip()
        return f"{tox_ini}\n"


def get_updated_pyproject_toml(
    data: str, ignore: Iterable[str] = (), all_extra_name: str = ""
) -> str:
    """
    Return the contents of a **pyproject.toml** file, updated to reflect the
    currently installed project versions, excluding those specified in
    `ignore`.

    Parameters:

    - data (str): The contents of a **pyproject.toml** file
    - ignore ([str]): One or more project names to leave as-is
    """
    ignore_set: Set[str] = _normalize_ignore_argument(ignore)

    def get_updated_requirement_string(requirement: str) -> str:
        return _get_updated_requirement_string(requirement, ignore=ignore_set)

    # Parse
    pyproject: Dict[str, Any] = tomli.loads(data)
    if ("build-system" in pyproject) and (
        "requires" in pyproject["build-system"]
    ):
        pyproject["build-system"]["requires"] = list(
            map(
                get_updated_requirement_string,
                pyproject["build-system"]["requires"],
            )
        )
        return tomli_w.dumps(pyproject)
    return data


def _update(
    path: str, ignore: Iterable[str] = (), all_extra_name: str = ""
) -> None:
    data: str
    update_function: Callable[[str], str]
    kwargs: Dict[str, Union[str, Iterable[str]]] = {}
    base_file_name: str = os.path.basename(path).lower()
    if base_file_name == "setup.cfg":
        update_function = get_updated_setup_cfg
        if all_extra_name:
            kwargs["all_extra_name"] = all_extra_name
    elif base_file_name == "pyproject.toml":
        update_function = get_updated_pyproject_toml
    elif base_file_name == "tox.ini":
        update_function = get_updated_tox_ini
    else:
        update_function = get_updated_requirements_txt
    kwargs["ignore"] = ignore
    file_io: IO[str]
    with open(path) as file_io:
        data = file_io.read()
    updated_data: str = update_function(data, **kwargs)
    if updated_data == data:
        print(f"All requirements were already up-to-date in {path}")
    else:
        print(f"Updating requirements in {path}")
        with open(path, "w") as file_io:
            file_io.write(updated_data)


def update(
    paths: Iterable[str],
    ignore: Iterable[str] = (),
    all_extra_name: str = "",
) -> None:
    """
    Update requirement versions in the specified files.

    Parameters:

    - path (str|[str}): One or more local paths to setup.cfg and/or
      requirements.txt files
    - ignore ([str]): One or more project names to ignore (leave as-is)
    - all_extra_name (str): If provided, an extra which consolidates
      the requirements for all other extras will be added/updated to
      *setup.cfg* (this argument is ignored for *requirements.txt* files)
    """
    if isinstance(paths, str):
        paths = (paths,)

    def update_(path: str) -> None:
        _update(path, ignore=ignore, all_extra_name=all_extra_name)

    list(map(update_, paths))


def main() -> None:
    parser: argparse.ArgumentParser = argparse.ArgumentParser(
        prog="daves-dev-tools requirements update",
        description=(
            "Update requirement versions in the specified files "
            "to align with currently installed versions of each distribution."
        ),
    )
    parser.add_argument(
        "-i",
        "--ignore",
        default=[],
        type=str,
        action="append",
        help=(
            "A comma-separated list of distributions to ignore (leave "
            "any requirements pertaining to the package as-is) "
        ),
    )
    parser.add_argument(
        "-aen",
        "--all-extra-name",
        default="",
        type=str,
        help=(
            "If provided, an extra which consolidates the requirements "
            "for all other extras will be added/updated to *setup.cfg* (this "
            "argument is ignored for *requirements.txt* files)"
        ),
    )
    parser.add_argument(
        "path",
        nargs="+",
        type=str,
        help=(
            "One or more local paths to a *setup.cfg* and/or "
            "*requirements.txt* file"
        ),
    )
    arguments: argparse.Namespace = parser.parse_args()
    update(
        paths=arguments.path,
        ignore=tuple(iter_parse_delimited_values(arguments.ignore)),
        all_extra_name=arguments.all_extra_name,
    )


if __name__ == "__main__":
    main()



import sys
from importlib import import_module
from types import ModuleType
from . import __name__ as _module_name
from .errors import get_exception_text


def _print_help() -> None:
    print(
        "Usage:\n"
        "  daves-dev-tools <command> [options]\n\n"
        "Commands:\n"
        "  git                         Tools for interacting with git repos.\n"
        "  requirements                Tools for managing requirement "
        "versions.\n"
        "  install-editable            Auto-discover and install packages "
        'in "develop"\n'
        "                              mode used by the current environment "
        "or by one or\n"
        "                              more specified packages/projects.\n"
        "  make-typed                  Configure your project to indicate to "
        "`mypy` that\n"
        "                              the contents are type-hinted.\n"
        "  uninstall-all               Uninstall all distributions from the "
        "current\n"
        "                              python environment except for those "
        "explicitly\n"
        "                              specified (and their dependencies).\n"
        "  clean                       Delete files from your project which "
        "are ignored\n"
        "                              by `git`, excepting files matching "
        "specified\n"
        "                              exclusion patterns.\n"
        "  distribute                  Build and distribute to PYPI (or other "
        "indexes)\n"
        "                              with one command."
    )


def _get_command() -> str:
    command: str = ""
    if len(sys.argv) > 1:
        command = sys.argv.pop(1).lower().replace("-", "_")
    return command


def main() -> None:
    """
    Run a sub-module `main` function.
    """
    command = _get_command()
    module: ModuleType
    try:
        try:
            module = import_module(f"{_module_name}.{command}.__main__")
        except ImportError:
            module = import_module(f"{_module_name}.{command}")
        module.main()  # type: ignore
    except Exception:
        print(get_exception_text())
        _print_help()


if __name__ == "__main__":
    main()



import argparse
import sys
from pipes import quote
from itertools import chain
from typing import Iterable, Tuple
from .requirements.utilities import (
    get_installed_distributions,
    get_requirements_required_distribution_names,
)
from .utilities import run


def uninstall_all(exclude: Iterable[str] = (), dry_run: bool = False) -> None:
    """
    Uninstall all distributions except for those requirementS specified
    in `exclude`.

    Parameters:

    - exclude ([str]): One or more requirement specifiers (for example:
      "requirement-name[extra-a,extra-b]" or ".[extra-a, extra-b]) and/or paths
      to a setup.cfg, pyproject.toml, tox.ini or requirements.txt file
    """
    name: str
    uninstall_distribution_names: Tuple[str, ...] = tuple(
        sorted(
            (
                set(get_installed_distributions().keys())
                - get_requirements_required_distribution_names(
                    chain(
                        ("pip", "setuptools", "wheel", "distribute"), exclude
                    )
                )
            ),
            key=lambda name: name.lower(),
        )
    )
    if uninstall_distribution_names:
        command: Tuple[str, ...] = (
            sys.executable,
            "-m",
            "pip",
            "uninstall",
            "-y",
        ) + uninstall_distribution_names
        if dry_run:
            print(" ".join(map(quote, command)))
        else:
            run(command)
    else:
        print("# No distributions found to uninstall")


def main() -> None:
    parser: argparse.ArgumentParser = argparse.ArgumentParser(
        prog="daves-dev-tools uninstall-all",
        description=(
            "This command will uninstall all distributions "
            "installed in the same environment as that from which this "
            "command is executed, excluding any specified by "
            "`-e EXCLUDE`"
        ),
    )
    parser.add_argument(
        "-e",
        "--exclude",
        default=[],
        type=str,
        action="append",
        help=(
            "One or more distribution specifiers, requirement files, "
            "setup.cfg files, pyproject.toml files, or tox.ini files "
            "denoting packages to exclude (along with all of their "
            "requirements) from those distributions to be "
            "uninstalled"
        ),
    )
    parser.add_argument(
        "-dr",
        "--dry-run",
        default=False,
        const=True,
        action="store_const",
        help=(
            "Print, but do not execute, the assembled `pip uninstall` command "
            "which, absent this flag, would be executed"
        ),
    )
    arguments: argparse.Namespace = parser.parse_args()
    uninstall_all(exclude=arguments.exclude, dry_run=arguments.dry_run)


if __name__ == "__main__":
    main()



import functools
import runpy
import sys
import os
from collections import deque
from shlex import quote
from itertools import chain
from subprocess import check_output
from urllib.parse import urlparse, urlunparse, ParseResult, quote as _quote
from typing import (
    Any,
    Callable,
    List,
    Iterable,
    Sequence,
    NoReturn,
    Optional,
    Union,
    Set,
    overload,
)

__all__: List[str] = [
    "lru_cache",
    "run",
    "iter_parse_delimited_values",
    "iter_sys_argv_pop",
    "sys_argv_pop",
    "iter_sys_argv_get",
    "sys_argv_get",
    "update_url_user_password",
]
lru_cache: Callable[..., Any] = functools.lru_cache


def update_url_user_password(
    url: str,
    user: str,
    password: str = "",
    quote: Callable[[str], str] = _quote,
) -> str:
    """
    Update a URL's user and password and return the result.

    Parameters:

    - url (str)
    - user (str)
    - password (str) = "": (optional)
    - quote = urllib.parse.quote: A function to use for escaping
      invalid character (defaults to `urllib.parse.quote`)
    """
    assert url and user
    parse_result: ParseResult = urlparse(url)
    host: str = parse_result.netloc.rpartition("@")[-1]
    user_password: str = quote(user)
    if password:
        user_password = f"{user_password}:{quote(password)}"
    return urlunparse(
        (
            parse_result.scheme,
            f"{user_password}@{host}",
            parse_result.path,
            parse_result.params,
            parse_result.query,
            parse_result.fragment,
        )
    )


def _iter_parse_delimited_value(value: str, delimiter: str) -> Iterable[str]:
    return value.split(delimiter)


def iter_parse_delimited_values(
    values: Iterable[str], delimiter: str = ","
) -> Iterable[str]:
    """
    This function iterates over input values which have been provided as a
    list or iterable and/or a single string of character-delimited values.
    A typical use-case is parsing multi-value command-line arguments.
    """
    if isinstance(values, str):
        values = (values,)

    def iter_parse_delimited_value_(value: str) -> Iterable[str]:
        return _iter_parse_delimited_value(value, delimiter=delimiter)

    return chain(*map(iter_parse_delimited_value_, values))


def run(command: Sequence[str], echo: bool = True) -> str:
    """
    This function runs a shell command, raises an error if a non-zero
    exit code is returned, and echo's both the command and output *if*
    the `echo` parameter is `True`.

    Parameters:

    - command (str|[str]): A shell command
    - echo (bool) = True: If `True`, the command and the output from the
      command will be printed to stdout
    """
    if echo:
        command_str: str
        if isinstance(command, str):
            command_str = command
        else:
            command_str = " ".join(map(quote, command))
        print(command_str)
    output: str = check_output(
        command,
        encoding="utf-8",
        universal_newlines=True,
        shell=isinstance(command, str),
    ).strip()
    if echo:
        print(output)
    return output


def _dummy_sys_exit(__status: object) -> None:
    return


def run_module_as_main(
    module_name: str,
    arguments: Sequence[str] = (),
    directory: str = ".",
    echo: bool = False,
) -> None:
    """
    This function runs a module as a main entry point, effectively as a CLI,
    but in the same sub-process as the calling function (thereby retaining
    all privileges granted to the current sub-process).

    Parameters:

    - module_name (str)
    - arguments ([str]) = (): The system arguments to pass to this command,
      replacing `sys.argv` while running the module).
    - directory (str) = ".": The directory in which the command should
      be executed (replacing `os.path.curdir` while running the module).
    - echo (bool) = False: If `True`, an equivalent shell command is printed
      to sys.stdout.
    """
    prior_sys_exit: Callable[[object], NoReturn] = sys.exit
    prior_sys_argv: List[str] = sys.argv
    if not isinstance(arguments, list):
        arguments = list(arguments)
    command_sys_argv: List[str] = sys.argv[:1] + arguments
    prior_current_directory: str = os.path.abspath(os.path.curdir)
    os.chdir(directory)
    try:
        if echo:
            print(
                " ".join(
                    map(
                        quote,
                        [sys.executable, "-m", module_name] + arguments,
                    )
                )
            )
        # Plugging a dummy function into `sys.exit` is necessary to avoid CLI
        # tools such as pip from ending the current process
        sys.exit = _dummy_sys_exit  # type: ignore
        sys.argv = command_sys_argv
        runpy.run_module(module_name, run_name="__main__")
    finally:
        sys.exit = prior_sys_exit  # type: ignore
        os.chdir(prior_current_directory)
        sys.argv = prior_sys_argv


def _validate_key(key: str) -> None:
    if not key.startswith("-"):
        raise ValueError(
            f'{repr(key)} is not a valid keyword. Keyword must begin with "-".'
        )


def _get_keys_set(keys: Optional[Iterable[str]] = None) -> Optional[Set[str]]:
    if keys is not None:
        if isinstance(keys, str):
            keys = {keys}
        elif not isinstance(keys, set):
            keys = set(keys)
        deque(map(_validate_key, keys), maxlen=0)
    return keys


def _iter_reversed_sys_argv_indices(
    keys: Optional[Iterable[str]] = None,
    argv: Optional[List[str]] = None,
    depth: int = 1,
) -> Iterable[int]:
    """
    In reverse order, yield the indices of the item in `sys.argv`
    index one of the indicated keys. If no `keys` are provided, yield the
    indices of all positional arguments.
    """
    if argv is None:
        argv = sys.argv
    keys = _get_keys_set(keys)
    length: int = len(argv)
    index: int
    negative_index: int
    for negative_index, value in enumerate(reversed(argv[depth:]), 1):
        index = length - negative_index
        if keys is not None:
            if value in keys:
                yield index
        elif not value.startswith("-"):
            try:
                if not argv[index - 1].startswith("-"):
                    yield index
            except IndexError:
                yield index


@overload
def iter_sys_argv_pop(
    keys: None,
    argv: Optional[List[str]],
    flag: bool,
) -> Iterable[str]:
    ...


@overload
def iter_sys_argv_pop(keys: None, argv: Optional[List[str]]) -> Iterable[str]:
    ...


@overload
def iter_sys_argv_pop(
    keys: None,
) -> Iterable[str]:
    ...


@overload
def iter_sys_argv_pop() -> Iterable[str]:
    ...


@overload
def iter_sys_argv_pop(
    keys: Optional[Iterable[str]],
    argv: Optional[List[str]],
    flag: Optional[bool],
    depth: int = 1,
) -> Iterable[Union[str, bool]]:
    ...


def iter_sys_argv_pop(
    keys: Optional[Iterable[str]] = None,
    argv: Optional[List[str]] = None,
    flag: Optional[bool] = None,
    depth: int = 1,
) -> Iterable[Union[str, bool]]:
    """
    Remove and yield all values, in reverse order, for an argument,
    from `sys.argv`.

    Parameters:

    - keys ([str]|str): Zero or more keywords/flags. If not `keys` are
      provided, this will yield positional argument values instead.
    - default (str|None) = None: If the argument is not found, this value
      is returned.
    - arg ([str]) = sys.argv: If provided, this list will be parsed instead
      of `sys.argv`.
    - flag (bool) = False: If `True`, treat the argument as a flag.
      If `False`, treat the argument as a flag only if it is not followed
      by a value.
    - depth (int) = 1: The number of items in `sys.argv` which should be
      interpreted as commands rather than positional arguments

    Examples:

    >>> deque(
    ...     map(
    ...         print,
    ...         iter_sys_argv_pop(
    ...             keys=("-r", "--requirement"),
    ...             argv=[
    ...                 "pip",
    ...                 "install",
    ...                 "-r", "requirements.txt",
    ...                 "--requirement", "dev_requirements.txt",
    ...                 "pytest",
    ...                 "tox",
    ...                 ".",
    ...             ]
    ...         )
    ...     ),
    ...     maxlen=0
    ... )
    dev_requirements.txt
    requirements.txt
    deque([], maxlen=0)

    >>> deque(
    ...     map(
    ...         print,
    ...         iter_sys_argv_pop(
    ...             argv=[
    ...                 "pip",
    ...                 "install",
    ...                 "-r", "requirements.txt",
    ...                 "--requirement", "dev_requirements.txt",
    ...                 "pytest",
    ...                 "tox",
    ...                 ".",
    ...             ],
    ...             depth=2,
    ...         )
    ...     ),
    ...     maxlen=0
    ... )
    .
    tox
    pytest
    deque([], maxlen=0)
    """
    return _iter_sys_argv_function(
        keys=keys, argv=argv, flag=flag, function=list.pop, depth=depth
    )


@overload
def iter_sys_argv_get(
    keys: None,
    argv: Optional[List[str]],
    flag: bool,
    depth: int,
) -> Iterable[str]:
    ...


@overload
def iter_sys_argv_get(
    keys: None, argv: Optional[List[str]], flag: Optional[bool], depth: int
) -> Iterable[str]:
    ...


@overload
def iter_sys_argv_get(
    keys: None,
) -> Iterable[str]:
    ...


@overload
def iter_sys_argv_get(
    keys: Optional[Iterable[str]],
    argv: Optional[List[str]],
    flag: Optional[bool],
    depth: int,
) -> Iterable[Union[str, bool]]:
    ...


def iter_sys_argv_get(
    keys: Optional[Iterable[str]] = None,
    argv: Optional[List[str]] = None,
    flag: Optional[bool] = None,
    depth: int = 1,
) -> Iterable[Union[str, bool]]:
    """
    Yield all values, in reverse order, for an argument, from `sys.argv`.
    If `keys` are provided, yield values for that keyword argument. If
    `keys` are *not* provided, yield positional arguments values.

    Parameters:

    - keys ([str]|str): Zero or more keywords/flags. If not `keys` are
      provided, this will yield positional argument values instead.
    - default (str|None) = None: If the argument is not found, this value
      is returned.
    - arg ([str]) = sys.argv: If provided, this list will be parsed instead
      of `sys.argv`.
    - flag (bool) = False: If `True`, treat the argument as a flag.
      If `False`, treat the argument as a flag only if it is not followed
      by a value.
    - depth (int) = 1: The number of items in `sys.argv` which should be
      interpreted as commands rather than positional arguments

    Examples:

    >>> deque(
    ...     map(
    ...         print,
    ...         iter_sys_argv_get(
    ...             keys=("-r", "--requirement"),
    ...             argv=[
    ...                 "pip",
    ...                 "install",
    ...                 "-r", "requirements.txt",
    ...                 "--requirement", "dev_requirements.txt",
    ...                 "pytest",
    ...                 "tox",
    ...                 ".",
    ...             ]
    ...         )
    ...     ),
    ...     maxlen=0
    ... )
    dev_requirements.txt
    requirements.txt
    deque([], maxlen=0)

    >>> deque(
    ...     map(
    ...         print,
    ...         iter_sys_argv_get(
    ...             argv=[
    ...                 "pip",
    ...                 "install",
    ...                 "-r", "requirements.txt",
    ...                 "--requirement", "dev_requirements.txt",
    ...                 "pytest",
    ...                 "tox",
    ...                 ".",
    ...             ],
    ...             depth=2,
    ...         )
    ...     ),
    ...     maxlen=0
    ... )
    .
    tox
    pytest
    deque([], maxlen=0)
    """
    return _iter_sys_argv_function(
        keys=keys, argv=argv, flag=flag, function=list.__getitem__, depth=depth
    )


def _iter_sys_argv_function(
    keys: Optional[Iterable[str]] = None,
    argv: Optional[List[str]] = None,
    flag: Optional[bool] = None,
    function: Callable[[list, int], Union[str, bool]] = list.pop,
    depth: int = 1,
) -> Iterable[Union[str, bool]]:
    if argv is None:
        argv = sys.argv
    index: Optional[int]
    for index in _iter_reversed_sys_argv_indices(
        keys=keys, argv=argv, depth=depth
    ):
        if keys is None:
            yield function(argv, index)
        else:
            value_index: int = index + 1
            # If there is no value following the keyword, it's a boolean flag
            if len(argv) <= value_index or argv[value_index].startswith("-"):
                assert flag is not False
                function(argv, index)
                yield True
            else:
                if flag:
                    value = True
                else:
                    value = function(argv, value_index)  # type: ignore
                function(argv, index)
                yield value


@overload
def sys_argv_pop(keys: None) -> Optional[str]:
    ...


@overload
def sys_argv_pop(
    keys: Optional[Iterable[str]] = None,
    default: Optional[str] = None,
    argv: Optional[List[str]] = None,
    flag: Optional[bool] = None,
) -> Union[str, bool, None]:
    ...


def sys_argv_pop(
    keys: Optional[Iterable[str]] = None,
    default: Optional[str] = None,
    argv: Optional[List[str]] = None,
    flag: Optional[bool] = None,
    depth: int = 1,
) -> Union[str, bool, None]:
    """
    Remove and return the last value for a keyword argument from `sys.argv`,
    or `None` if there are not any positional arguments.

    Parameters:

    - keys ([str]|str|None): Zero or more keywords. For example:
      `("-u", "--user")` or `"--user"`. If no `keys` are provided, the last
      positional argument will be removed and returned instead.
    - default (str|None) = None: If the argument is not found, this value
      is returned.
    - arg ([str]) = sys.argv: If provided, this list will be parsed instead
      of `sys.argv`.
    - depth (int) = 1: The number of items in `sys.argv` which should be
      interpreted as commands rather than positional arguments

    Examples:

    >>> print(sys_argv_pop(
    ...     keys=("-r", "--requirement"),
    ...     argv=[
    ...         "pip",
    ...         "install",
    ...         "-r", "requirements.txt",
    ...         "--requirement", "dev_requirements.txt",
    ...         "pytest",
    ...         "tox",
    ...         ".",
    ...     ]
    ... ))
    dev_requirements.txt

    >>> print(sys_argv_pop(
    ...     argv=[
    ...         "pip",
    ...         "install",
    ...         "-r", "requirements.txt",
    ...         "--requirement", "dev_requirements.txt",
    ...         "pytest",
    ...         "tox",
    ...         ".",
    ...     ]
    ... ))
    .
    """
    try:
        return next(
            iter(
                iter_sys_argv_pop(keys=keys, argv=argv, flag=flag, depth=depth)
            )
        )
    except StopIteration:
        return default


@overload
def sys_argv_get(keys: None) -> Optional[str]:
    ...


@overload
def sys_argv_get(
    keys: Optional[Iterable[str]] = None,
    default: Optional[str] = None,
    argv: Optional[List[str]] = None,
    flag: Optional[bool] = None,
) -> Union[str, bool, None]:
    ...


def sys_argv_get(
    keys: Optional[Iterable[str]] = None,
    default: Optional[str] = None,
    argv: Optional[List[str]] = None,
    flag: Optional[bool] = None,
    depth: int = 1,
) -> Union[str, bool, None]:
    """
    If `keys` are provided: Return the last value for a keyword argument from
    `sys.argv`, or `None` if the keyword argument is not present.

    If `keys` are *not* provided, return the last positional argument or
    `None` if there are no positional arguments.

    Parameters:

    - keys ([str]|str|None): Zero or more keywords. For example:
      `("-u", "--user")` or `"--user"`. If no `keys` are provided, the last
      positional argument will be removed and returned instead.
    - default (str|None) = None: If the argument is not found, this value
      is returned.
    - arg ([str]) = sys.argv: If provided, this list will be parsed instead
      of `sys.argv`.

    Examples:

    >>> print(sys_argv_get(
    ...     keys=("-r", "--requirement"),
    ...     argv=[
    ...         "pip",
    ...         "install",
    ...         "-r", "requirements.txt",
    ...         "--requirement", "dev_requirements.txt",
    ...         "pytest",
    ...         "tox",
    ...         ".",
    ...     ]
    ... ))
    dev_requirements.txt

    >>> print(sys_argv_get(
    ...     argv=[
    ...         "pip",
    ...         "install",
    ...         "-r", "requirements.txt",
    ...         "--requirement", "dev_requirements.txt",
    ...         "pytest",
    ...         "tox",
    ...         ".",
    ...     ]
    ... ))
    .
    """
    try:
        return next(
            iter(
                iter_sys_argv_get(keys=keys, argv=argv, flag=flag, depth=depth)
            )
        )
    except StopIteration:
        return default



import pkg_resources
import argparse
from fnmatch import fnmatch
from itertools import chain
from typing import Iterable, Tuple, Set
from more_itertools import unique_everseen
from .utilities import (
    get_required_distribution_names,
    get_distribution,
    install_requirement,
    iter_configuration_file_requirement_strings,
    get_requirement_string_distribution_name,
    normalize_name,
    is_configuration_file,
)
from ..utilities import iter_parse_delimited_values

_DO_NOT_PIN_DISTRIBUTION_NAMES: Set[str] = {
    # standard library
    "importlib-metadata",
    "importlib-resources",
}


def get_frozen_requirements(
    requirements: Iterable[str] = (),
    exclude: Iterable[str] = (),
    exclude_recursive: Iterable[str] = (),
    no_version: Iterable[str] = (),
) -> Tuple[str, ...]:
    """
    Get the (frozen) requirements for one or more specified distributions or
    configuration files.

    Parameters:

    - requirements ([str]): One or more requirement specifiers (for example:
      "requirement-name[extra-a,extra-b]" or ".[extra-a, extra-b]) and/or paths
      to a setup.cfg, pyproject.toml, tox.ini or requirements.txt file
    - exclude ([str]): One or more distributions to exclude/ignore
    - exclude_recursive ([str]): One or more distributions to exclude/ignore.
      Note: Excluding a distribution here excludes all requirements which would
      be identified through recursively.
      those requirements occur elsewhere.
    - no_version ([str]) = (): Exclude version numbers from the output
      (only return distribution names)
    """
    # Separate requirement strings from requirement files
    if isinstance(requirements, str):
        requirements = {requirements}
    else:
        requirements = set(requirements)
    if isinstance(no_version, str):
        no_version = (no_version,)
    elif not isinstance(no_version, tuple):
        no_version = tuple(no_version)
    requirement_files: Set[str] = set(
        filter(is_configuration_file, requirements)
    )
    requirement_strings: Set[str] = requirements - requirement_files
    name: str
    return tuple(
        sorted(
            _iter_frozen_requirements(
                unique_everseen(
                    chain(
                        requirement_strings,
                        *map(
                            iter_configuration_file_requirement_strings,
                            requirement_files,
                        ),
                    )
                ),
                exclude=set(
                    chain(
                        # Exclude requirement strings which are *not*
                        # distribution names (such as editable package paths),
                        # as in these cases we are typically looking for this
                        # package's dependencies
                        (
                            set(
                                map(
                                    get_requirement_string_distribution_name,
                                    requirement_strings,
                                )
                            )
                            - set(map(normalize_name, requirement_strings))
                        ),
                        map(normalize_name, exclude),
                    )
                ),
                exclude_recursive=set(map(normalize_name, exclude_recursive)),
                no_version=no_version,
            ),
            key=lambda name: name.lower(),
        )
    )


def _iter_frozen_requirements(
    requirement_strings: Iterable[str],
    exclude: Set[str],
    exclude_recursive: Set[str],
    no_version: Iterable[str] = (),
) -> Iterable[str]:
    def get_requirement_string(distribution_name: str) -> str:
        def distribution_name_matches_pattern(pattern: str) -> bool:
            return fnmatch(distribution_name, pattern)

        # * Don't pin importlib-metadata, as it is part of the standard
        #   library so we should use the version distributed with
        #   python, and...
        # * Only include the version in the requirement string if
        #   the package name does not match any patterns provided in the
        #   `no_version` argument
        if (distribution_name in _DO_NOT_PIN_DISTRIBUTION_NAMES) or any(
            map(distribution_name_matches_pattern, no_version)
        ):
            return distribution_name
        distribution: pkg_resources.Distribution
        try:
            distribution = get_distribution(distribution_name)
        except KeyError:
            # If the distribution is missing, install it
            install_requirement(distribution_name, echo=False)
            distribution = get_distribution(distribution_name)
        requirement_string: str = str(distribution.as_requirement())
        # Only include dataclasses for python 3.6
        if distribution_name == "dataclasses":
            requirement_string = (
                f'{requirement_string}; python_version == "3.6"'
            )
        return requirement_string

    def get_required_distribution_names_(requirement_string: str) -> Set[str]:
        name: str = get_requirement_string_distribution_name(
            requirement_string
        )
        if name in exclude_recursive:
            return set()
        return (
            get_required_distribution_names(
                requirement_string, exclude=exclude_recursive
            )
            | {name}
        ) - exclude

    requirements: Iterable[str] = unique_everseen(
        chain(*map(get_required_distribution_names_, requirement_strings)),
    )

    requirements = map(get_requirement_string, requirements)
    return requirements


def freeze(
    requirements: Iterable[str] = (),
    exclude: Iterable[str] = (),
    exclude_recursive: Iterable[str] = (),
    no_version: Iterable[str] = (),
) -> None:
    """
    Print the (frozen) requirements for one or more specified requirements or
    configuration files.

    Parameters:

    - requirements ([str]): One or more requirement specifiers (for example:
      "requirement-name[extra-a,extra-b]" or ".[extra-a, extra-b]) and/or paths
      to a setup.py, setup.cfg, pyproject.toml, tox.ini or requirements.txt
      file
    - exclude ([str]): One or more distributions to exclude/ignore
    - exclude_recursive ([str]): One or more distributions to exclude/ignore.
      Note: Excluding a distribution here excludes all requirements which would
      be identified through recursively.
      those requirements occur elsewhere.
    - no_version ([str]) = (): Exclude version numbers from the output
      (only print distribution names) for package names matching any of these
      patterns
    """
    print(
        "\n".join(
            get_frozen_requirements(
                requirements=requirements,
                exclude=exclude,
                exclude_recursive=exclude_recursive,
                no_version=no_version,
            )
        )
    )


def main() -> None:
    parser: argparse.ArgumentParser = argparse.ArgumentParser(
        prog="daves-dev-tools requirements freeze",
        description=(
            "This command prints dependencies inferred from an installed "
            "distribution or project, in a similar format to the "
            "output of `pip freeze`, except that all generated requirements "
            'are specified in the format "distribution-name==0.0.0" '
            "(including for editable installations). Using this command "
            "instead of `pip freeze` to generate requirement files ensures "
            "that you don't bloat your requirements files with superfluous "
            "distributions."
        ),
    )
    parser.add_argument(
        "requirement",
        nargs="+",
        type=str,
        help=(
            "One or more requirement specifiers (for example: "
            '"requirement-name", "requirement-name[extra-a,extra-b]", '
            '".[extra-a, extra-b]" or '
            '"../other-editable-package-directory[extra-a, extra-b]) '
            "and/or paths to a setup.py, setup.cfg, pyproject.toml, "
            "tox.ini or requirements.txt file"
        ),
    )
    parser.add_argument(
        "-e",
        "--exclude",
        default=[],
        type=str,
        action="append",
        help=(
            "A distribution (or comma-separated list of distributions) to "
            "exclude from the output"
        ),
    )
    parser.add_argument(
        "-er",
        "--exclude-recursive",
        default=[],
        type=str,
        action="append",
        help=(
            "A distribution (or comma-separated list of distributions) to "
            "exclude from the output. Unlike -e / --exclude, "
            "this argument also precludes recursive requirement discovery "
            "for the specified packages, thereby excluding all of the "
            "excluded package's requirements which are not required by "
            "another (non-excluded) distribution."
        ),
    )
    parser.add_argument(
        "-nv",
        "--no-version",
        type=str,
        default=[],
        action="append",
        help=(
            "Don't include versions (only output distribution names) "
            "for packages matching this glob pattern (note: the value must "
            "be single-quoted if it contain wildcards)"
        ),
    )
    arguments: argparse.Namespace = parser.parse_args()
    freeze(
        requirements=arguments.requirement,
        exclude=tuple(iter_parse_delimited_values(arguments.exclude)),
        exclude_recursive=tuple(
            iter_parse_delimited_values(arguments.exclude_recursive)
        ),
        no_version=arguments.no_version,
    )


if __name__ == "__main__":
    main()



import sys
import os
import tomli
import pkg_resources
import importlib_metadata
from subprocess import check_output, CalledProcessError
from collections import deque
from warnings import warn
from configparser import ConfigParser, SectionProxy
from enum import Enum, auto
from itertools import chain
from typing import (
    Optional,
    Dict,
    Iterable,
    Set,
    Tuple,
    List,
    IO,
    Union,
    Callable,
    Any,
)
from packaging.utils import canonicalize_name
from packaging.requirements import InvalidRequirement, Requirement
from more_itertools import unique_everseen
from ..utilities import lru_cache, run
from ..errors import append_exception_text, get_exception_text

_BUILTIN_DISTRIBUTION_NAMES: Tuple[str] = ("distribute",)
# This variable tracks the absolute file paths from which a package has been
# re-installed, in order to avoid performing a reinstall redundantly
_reinstalled_locations: Set[str] = set()


@lru_cache()
def normalize_name(name: str) -> str:
    """
    Normalize a project/distribution name
    """
    return pkg_resources.safe_name(canonicalize_name(name)).lower()


class ConfigurationFileType(Enum):

    REQUIREMENTS_TXT = auto()
    SETUP_CFG = auto()
    TOX_INI = auto()
    PYPROJECT_TOML = auto()


@lru_cache()
def get_configuration_file_type(path: str) -> ConfigurationFileType:
    if not os.path.isfile(path):
        raise FileNotFoundError(path)
    basename: str = os.path.basename(path).lower()
    if basename == "setup.cfg":
        return ConfigurationFileType.SETUP_CFG
    elif basename == "tox.ini":
        return ConfigurationFileType.TOX_INI
    elif basename == "pyproject.toml":
        return ConfigurationFileType.PYPROJECT_TOML
    elif basename.endswith(".txt"):
        return ConfigurationFileType.REQUIREMENTS_TXT
    else:
        raise ValueError(
            f"{path} is not a recognized type of configuration file."
        )


def is_configuration_file(path: str) -> bool:
    try:
        get_configuration_file_type(path)
    except (FileNotFoundError, ValueError):
        return False
    return True


def refresh_working_set() -> None:
    """
    Force a refresh of all distribution information and clear related caches
    """
    get_installed_distributions.cache_clear()
    is_editable.cache_clear()
    is_installed.cache_clear()
    get_requirement_string_distribution_name.cache_clear()
    pkg_resources.working_set.entries = []
    pkg_resources.working_set.__init__()  # type: ignore


@lru_cache()
def get_installed_distributions() -> Dict[str, pkg_resources.Distribution]:
    """
    Return a dictionary of installed distributions.
    """
    installed: Dict[str, pkg_resources.Distribution] = {}
    for distribution in pkg_resources.working_set:
        if _distribution_is_editable(distribution):
            _reload_distribution_information(distribution)
        installed[normalize_name(distribution.project_name)] = distribution
    return installed


def get_distribution(name: str) -> pkg_resources.Distribution:
    return get_installed_distributions()[normalize_name(name)]


@lru_cache()
def is_installed(distribution_name: str) -> bool:
    return normalize_name(distribution_name) in get_installed_distributions()


def get_requirement_distribution_name(requirement: Requirement) -> str:
    return normalize_name(requirement.name)


@lru_cache()
def get_requirement_string_distribution_name(requirement_string: str) -> str:
    return get_requirement_distribution_name(
        get_requirement(requirement_string)
    )


@lru_cache()
def is_requirement_string(requirement_string: str) -> bool:
    try:
        Requirement(requirement_string)
    except InvalidRequirement:
        return False
    return True


def _iter_file_requirement_strings(path: str) -> Iterable[str]:
    lines: List[str]
    requirement_file_io: IO[str]
    with open(path) as requirement_file_io:
        lines = requirement_file_io.readlines()
    return filter(is_requirement_string, lines)


def _iter_setup_cfg_requirement_strings(path: str) -> Iterable[str]:
    parser: ConfigParser = ConfigParser()
    parser.read(path)
    requirement_strings: Iterable[str] = ()
    if ("options" in parser) and ("install_requires" in parser["options"]):
        requirement_strings = chain(
            requirement_strings,
            filter(
                is_requirement_string,
                parser["options"]["install_requires"].split("\n"),
            ),
        )
    if "options.extras_require" in parser:
        extras_require: SectionProxy = parser["options.extras_require"]
        extra_requirements_string: str
        for extra_requirements_string in extras_require.values():
            requirement_strings = chain(
                requirement_strings,
                filter(
                    is_requirement_string,
                    extra_requirements_string.split("\n"),
                ),
            )
    return unique_everseen(requirement_strings)


def _iter_tox_ini_requirement_strings(path: str) -> Iterable[str]:
    parser: ConfigParser = ConfigParser()
    parser.read(path)

    def get_section_option_requirements(
        section_name: str, option_name: str
    ) -> Iterable[str]:
        if parser.has_option(section_name, option_name):
            return filter(
                is_requirement_string,
                parser.get(section_name, option_name).split("\n"),
            )
        return ()

    def get_section_requirements(section_name: str) -> Iterable[str]:
        requirements: Iterable[str] = get_section_option_requirements(
            section_name, "deps"
        )
        if section_name == "tox":
            requirements = chain(
                requirements,
                get_section_option_requirements(section_name, "requires"),
            )
        return requirements

    return unique_everseen(
        chain(("tox",), *map(get_section_requirements, parser.sections()))
    )


def _iter_pyproject_toml_requirement_strings(path: str) -> Iterable[str]:
    pyproject_io: IO[str]
    with open(path) as pyproject_io:
        pyproject: Dict[str, Any] = tomli.loads(pyproject_io.read())
        if ("build-system" in pyproject) and (
            "requires" in pyproject["build-system"]
        ):
            return pyproject["build-system"]["requires"]
    return ()


def iter_configuration_file_requirement_strings(path: str) -> Iterable[str]:
    """
    Read a configuration file and yield the parsed requirements.
    """
    configuration_file_type: ConfigurationFileType = (
        get_configuration_file_type(path)
    )
    if configuration_file_type == ConfigurationFileType.SETUP_CFG:
        return _iter_setup_cfg_requirement_strings(path)
    elif configuration_file_type == ConfigurationFileType.PYPROJECT_TOML:
        return _iter_pyproject_toml_requirement_strings(path)
    elif configuration_file_type == ConfigurationFileType.TOX_INI:
        return _iter_tox_ini_requirement_strings(path)
    else:
        assert (
            configuration_file_type == ConfigurationFileType.REQUIREMENTS_TXT
        )
        return _iter_file_requirement_strings(path)


@lru_cache()
def is_editable(distribution_name: str) -> bool:
    """
    Return `True` if the indicated distribution is an editable installation.
    """
    return _distribution_is_editable(get_distribution(distribution_name))


def _distribution_is_editable(
    distribution: pkg_resources.Distribution,
) -> bool:
    """
    Return `True` if the `distribution` is an editable installation.
    """
    egg_link_file_name: str = f"{distribution.project_name}.egg-link"

    def project_egg_link_exists(path: str) -> bool:
        return os.path.isfile(os.path.join(path, egg_link_file_name))

    return any(map(project_egg_link_exists, sys.path))


def _get_setup_cfg_metadata(path: str, key: str) -> str:
    if os.path.basename(path).lower() != "setup.cfg":
        if not os.path.isdir(path):
            path = os.path.dirname(path)
        path = os.path.join(path, "setup.cfg")
    if os.path.isfile(path):
        parser: ConfigParser = ConfigParser()
        parser.read(path)
        if "metadata" in parser:
            return parser.get("metadata", key, fallback="")
        else:
            warn(f"No `metadata` section found in: {path}")
    return ""


def _get_setup_py_metadata(path: str, args: Tuple[str, ...]) -> str:
    """
    Execute a setup.py script with `args` and return the response.

    Parameters:

    - path (str)
    - args ([str])
    """
    value: str = ""
    current_directory: str = os.path.abspath(os.curdir)
    directory: str = path
    try:
        if os.path.basename(path).lower() == "setup.py":
            directory = os.path.dirname(path)
            os.chdir(directory)
        else:
            if not os.path.isdir(path):
                directory = os.path.dirname(path)
            os.chdir(directory)
            path = os.path.join(directory, "setup.py")
        if os.path.isfile(path):
            command: Tuple[str, ...] = (sys.executable, path) + args
            try:
                value = (
                    check_output(
                        command, encoding="utf-8", universal_newlines=True
                    )
                    .strip()
                    .split("\n")[-1]
                )
            except CalledProcessError:
                warn(
                    f"A package name could not be found in {path}, "
                    "attempting to refresh distribution and egg info"
                    f"\nError ignored: {get_exception_text()}"
                )
                # re-write distribution and egg information and attempt to
                # get the name again
                setup_dist_egg_info(directory)
                try:
                    value = (
                        check_output(
                            command, encoding="utf-8", universal_newlines=True
                        )
                        .strip()
                        .split("\n")[-1]
                    )
                except Exception:
                    warn(
                        f"A package name could not be found in {path}"
                        f"\nError ignored: {get_exception_text()}"
                    )
    finally:
        os.chdir(current_directory)
    return value


def get_setup_distribution_name(path: str) -> str:
    """
    Get a distribution's name from setup.py or setup.cfg
    """
    return normalize_name(
        _get_setup_cfg_metadata(path, "name")
        or _get_setup_py_metadata(path, ("--name",))
    )


def get_setup_distribution_version(path: str) -> str:
    """
    Get a distribution's version from setup.py or setup.cfg
    """
    return _get_setup_cfg_metadata(path, "version") or _get_setup_py_metadata(
        path, ("--version",)
    )


def _setup(arguments: Tuple[str, ...]) -> None:
    try:
        check_output((sys.executable, "setup.py") + arguments)
    except CalledProcessError:
        warn(f"Ignoring error: {get_exception_text()}")


def _setup_location(
    location: str, arguments: Iterable[Tuple[str, ...]]
) -> None:
    # If there is no setup.py file, we can't update egg info
    if not os.path.isfile(os.path.join(location, "setup.py")):
        return
    if isinstance(arguments, str):
        arguments = (arguments,)
    current_directory: str = os.path.abspath(os.curdir)
    os.chdir(location)
    try:
        deque(map(_setup, arguments), maxlen=0)
    finally:
        os.chdir(current_directory)


def setup_dist_egg_info(directory: str) -> None:
    """
    Refresh dist-info and egg-info for the editable package installed in
    `directory`
    """
    directory = os.path.abspath(directory)
    if not os.path.isdir(directory):
        directory = os.path.dirname(directory)
    _setup_location(directory, (("-q", "dist_info"), ("-q", "egg_info")))


def _reload_distribution_information(
    distribution: pkg_resources.Distribution,
) -> pkg_resources.Distribution:
    if _distribution_is_editable(distribution):
        setup_dist_egg_info(distribution.location)
    # Clear the distribution information cache
    for cached_property_name in (
        "_pkg_info",
        "_key",
        "_parsed_version",
        "_version",
        "__dep_map",
    ):
        try:
            delattr(distribution, cached_property_name)
        except AttributeError:
            pass
    return distribution


def reload_distribution_information(name: str) -> pkg_resources.Distribution:
    return _reload_distribution_information(get_distribution(name))


def setup_dist_info(directory: str) -> None:
    """
    Refresh dist-info and egg-info for the editable package installed in
    `directory`
    """
    directory = os.path.abspath(directory)
    if not os.path.isdir(directory):
        directory = os.path.dirname(directory)
    return _setup_location(directory, (("-q", "dist_info"),))


def setup_egg_info(directory: str) -> None:
    """
    Refresh dist-info and egg-info for the editable package installed in
    `directory`
    """
    directory = os.path.abspath(directory)
    if not os.path.isdir(directory):
        directory = os.path.dirname(directory)
    return _setup_location(directory, (("-q", "egg_info"),))


def _get_location_distribution_name(location: str) -> str:
    """
    Get a distribution name based on an installation location, or return
    an empty string if no distribution can be found
    """
    location = os.path.abspath(location)

    def _is_in_location(
        name_distribution: Tuple[str, pkg_resources.Distribution]
    ) -> bool:
        return os.path.abspath(name_distribution[1].location) == location

    def _get_name(
        name_distribution: Tuple[str, pkg_resources.Distribution]
    ) -> str:
        return name_distribution[0]

    try:
        return next(
            map(
                _get_name,
                filter(_is_in_location, get_installed_distributions().items()),
            )
        )
    except StopIteration:
        return get_setup_distribution_name(location)


def _get_pkg_requirement(
    requirement_string: str,
) -> pkg_resources.Requirement:
    requirement: Union[
        Requirement, pkg_resources.Requirement
    ] = _get_requirement(requirement_string, pkg_resources.Requirement.parse)
    assert isinstance(requirement, pkg_resources.Requirement)
    return requirement


def get_requirement(
    requirement_string: str,
) -> Requirement:
    requirement: Union[
        Requirement, pkg_resources.Requirement
    ] = _get_requirement(requirement_string, Requirement)
    assert isinstance(requirement, Requirement)
    return requirement


def _get_requirement(
    requirement_string: str,
    constructor: Callable[
        [str], Union[Requirement, pkg_resources.Requirement]
    ],
) -> Union[Requirement, pkg_resources.Requirement]:
    try:
        return constructor(requirement_string)
    except (
        InvalidRequirement,
        getattr(
            pkg_resources, "extern"
        ).packaging.requirements.InvalidRequirement,
        getattr(pkg_resources, "RequirementParseError"),
    ):
        # Try to parse the requirement as an installation target location,
        # such as can be used with `pip install`
        location: str = requirement_string
        extras: str = ""
        if "[" in requirement_string and requirement_string.endswith("]"):
            parts: List[str] = requirement_string.split("[")
            location = "[".join(parts[:-1])
            extras = f"[{parts[-1]}"
        location = os.path.abspath(location)
        name: str = _get_location_distribution_name(location)
        assert name, f"No distribution found in {location}"
        return constructor(f"{name}{extras}")


def get_required_distribution_names(
    requirement_string: str,
    exclude: Iterable[str] = (),
    recursive: bool = True,
    echo: bool = False,
) -> Set[str]:
    """
    Return a `set` of all distribution names which are required by the
    distribution specified in `requirement_string`.

    Parameters:

    - requirement_string (str): A distribution name, or a requirement string
      indicating both a distribution name and extras.
    - exclude ([str]): The name of one or more distributions to *exclude*
      from requirements lookup. Please note that excluding a distribution will
      also halt recursive lookup of requirements for that distribution.
    - recursive (bool): If `True` (the default), required distributions will
      be obtained recursively.
    - echo (bool) = False: If `True`, commands and responses executed in
      subprocesses will be printed to `sys.stdout`
    """
    if isinstance(exclude, str):
        exclude = {normalize_name(exclude)}
    else:
        exclude = set(map(normalize_name, exclude))
    return set(
        _iter_requirement_names(
            _get_pkg_requirement(requirement_string),
            exclude=exclude,
            recursive=recursive,
            echo=echo,
        )
    )


def _get_pkg_requirement_name(requirement: pkg_resources.Requirement) -> str:
    return normalize_name(requirement.project_name)


def install_requirement(
    requirement: Union[str, Requirement, pkg_resources.Requirement],
    echo: bool = True,
) -> None:
    """
    Install a requirement

    Parameters:

    - requirement (str)
    - echo (bool) = True: If `True` (default), the `pip install`
      commands will be echoed to `sys.stdout`
    """
    if isinstance(requirement, str):
        requirement = Requirement(requirement)
    return _install_requirement(requirement, echo=echo)


def _install_requirement_string(
    requirement_string: str,
    name: str = "",
    editable: bool = False,
    echo: bool = False,
) -> None:
    uncaught_error: Optional[Exception] = None
    flags: Tuple[str, ...]
    for flags in chain(
        (("--user",), ()),
        (
            (("--user", "--force-reinstall"), ("--force-reinstall",))
            if editable
            else ()
        ),
    ):
        try:
            run(
                (
                    (sys.executable, "-m", "pip", "install")
                    + flags
                    + (requirement_string,)
                ),
                echo=echo,
            )
            uncaught_error = None
            break
        except CalledProcessError as error:
            if (uncaught_error is None) or (not flags):
                uncaught_error = error
    if uncaught_error is not None:
        append_exception_text(
            uncaught_error,
            (
                f"\nCould not install {name}"
                if name == requirement_string
                else (f"\nCould not install {name} from {requirement_string}")
            )
            if name
            else (f"\nCould not install {requirement_string}"),
        )
        raise uncaught_error


def _install_requirement(
    requirement: Union[Requirement, pkg_resources.Requirement],
    echo: bool = True,
) -> None:
    requirement_string: str = str(requirement)
    # Get the distribution name
    name: str = normalize_name(
        requirement.name
        if isinstance(requirement, Requirement)
        else requirement.project_name
    )
    distribution: Optional[pkg_resources.Distribution] = None
    editable: bool = False
    try:
        distribution = get_distribution(name)
        editable = _distribution_is_editable(distribution)
    except KeyError:
        pass
    # If the requirement is installed and editable, re-install from
    # the editable location
    if distribution and editable:
        # Assemble a requirement specifier for the editable install
        requirement_string = distribution.location
        if requirement.extras:
            requirement_string = (
                f"{requirement_string}[{','.join(requirement.extras)}]"
            )
    _install_requirement_string(
        requirement_string=requirement_string,
        name=name,
        editable=editable,
        echo=echo,
    )
    # Refresh the metadata
    if distribution:
        _reload_distribution_information(distribution)
    else:
        refresh_working_set()


def _get_pkg_requirement_distribution(
    requirement: pkg_resources.Requirement,
    name: str,
    reinstall: bool = True,
    echo: bool = False,
) -> Optional[pkg_resources.Distribution]:
    if name in _BUILTIN_DISTRIBUTION_NAMES:
        return None
    try:
        return get_installed_distributions()[name]
    except KeyError:
        if not reinstall:
            raise
        if echo:
            warn(
                f'The required distribution "{name}" was not installed, '
                "attempting to install it now..."
            )
        # Attempt to install the requirement...
        install_requirement(requirement, echo=echo)
        return _get_pkg_requirement_distribution(
            requirement, name, reinstall=False, echo=echo
        )


def _iter_requirement_names(
    requirement: pkg_resources.Requirement,
    exclude: Set[str],
    recursive: bool = True,
    echo: bool = False,
) -> Iterable[str]:
    name: str = normalize_name(requirement.project_name)
    extras: Set[str] = set(map(normalize_name, requirement.extras))
    if name in exclude:
        return ()
    # Ensure we don't follow the same requirement again, causing cyclic
    # recursion
    exclude.add(name)
    distribution: Optional[
        pkg_resources.Distribution
    ] = _get_pkg_requirement_distribution(requirement, name, echo=echo)
    if distribution is None:
        return ()
    requirements: List[pkg_resources.Requirement] = distribution.requires(
        extras=tuple(sorted(extras))
    )
    lateral_exclude: Set[str] = set()

    def iter_requirement_names_(
        requirement_: pkg_resources.Requirement,
    ) -> Iterable[str]:
        return _iter_requirement_names(
            requirement_,
            exclude=(
                exclude
                | (lateral_exclude - {_get_pkg_requirement_name(requirement_)})
            ),
            recursive=recursive,
            echo=echo,
        )

    def not_excluded(name: str) -> bool:
        if name not in exclude:
            # Add this to the exclusions
            lateral_exclude.add(name)
            return True
        return False

    if recursive:
        requirement_names = chain(
            filter(not_excluded, map(_get_pkg_requirement_name, requirements)),
            *map(iter_requirement_names_, requirements),
        )
    return requirement_names


def _iter_requirement_strings_required_distribution_names(
    requirement_strings: Iterable[str],
    echo: bool = False,
) -> Iterable[str]:
    visited_requirement_strings: Set[str] = set()
    if isinstance(requirement_strings, str):
        requirement_strings = (requirement_strings,)

    def get_required_distribution_names_(requirement_string: str) -> Set[str]:
        if requirement_string not in visited_requirement_strings:
            try:
                name: str = get_requirement_string_distribution_name(
                    requirement_string
                )
                visited_requirement_strings.add(requirement_string)
                return get_required_distribution_names(
                    requirement_string, echo=echo
                ) | {name}
            except KeyError:
                pass
        return set()

    return unique_everseen(
        chain(*map(get_required_distribution_names_, requirement_strings)),
    )


def get_requirements_required_distribution_names(
    requirements: Iterable[str] = (),
    echo: bool = False,
) -> Set[str]:
    """
    Get the distributions required by one or more specified distributions or
    configuration files.

    Parameters:

    - requirements ([str]): One or more requirement specifiers (for example:
      "requirement-name[extra-a,extra-b]" or ".[extra-a, extra-b]) and/or paths
      to a setup.cfg, pyproject.toml, tox.ini or requirements.txt file
    """
    # Separate requirement strings from requirement files
    if isinstance(requirements, str):
        requirements = {requirements}
    else:
        requirements = set(requirements)
    requirement_files: Set[str] = set(
        filter(is_configuration_file, requirements)
    )
    requirement_strings: Set[str] = requirements - requirement_files
    name: str
    return set(
        sorted(
            _iter_requirement_strings_required_distribution_names(
                unique_everseen(
                    chain(
                        requirement_strings,
                        *map(
                            iter_configuration_file_requirement_strings,
                            requirement_files,
                        ),
                    )
                ),
                echo=echo,
            ),
            key=lambda name: name.lower(),
        )
    )


def iter_distribution_location_file_paths(location: str) -> Iterable[str]:
    location = os.path.abspath(location)
    name: str = get_setup_distribution_name(location)
    setup_dist_egg_info(location)
    metadata_path: str = os.path.join(
        location, f"{pkg_resources.to_filename(name)}.egg-info"
    )
    distribution: importlib_metadata.Distribution = (
        importlib_metadata.Distribution.at(metadata_path)
    )
    if not distribution.files:
        raise RuntimeError(f"No metadata found at {metadata_path}")
    return map(os.path.abspath, distribution.files)



import sys
from traceback import format_exception
from typing import Any, List


def get_exception_text() -> str:
    """
    When called within an exception, this function returns a text
    representation of the error matching what is found in
    `traceback.print_exception`, but is returned as a string value rather than
    printing.
    """
    return "".join(format_exception(*sys.exc_info()))


def append_exception_text(error: Exception, message: str) -> None:
    """
    Cause `message` to be appended to an error's exception text.
    """
    last_attribute_name: str
    for last_attribute_name in ("strerror", "msg"):
        last_attribute_value = getattr(error, last_attribute_name, "")
        if last_attribute_value:
            setattr(
                error, last_attribute_name, f"{last_attribute_value}{message}"
            )
            break
    if not last_attribute_value:
        index: int
        arg: Any
        reversed_args: List[Any] = list(reversed(error.args)) or [""]
        for index, value in enumerate(reversed_args):
            if isinstance(value, str):
                reversed_args[index] = f"{value}{message}"
                break
        error.args = tuple(reversed(reversed_args))



import argparse
import os
from subprocess import check_call
from tempfile import mkdtemp
from itertools import chain
from shutil import rmtree
from typing import Iterable, List, Tuple, Iterator
from glob import iglob
from ..cerberus import apply_cerberus_path_arguments
from ..utilities import update_url_user_password


def _iglob_recursive(pathname: str) -> Iterator[str]:
    return iglob(pathname, recursive=True)


@apply_cerberus_path_arguments(
    "cerberus_url",
    user="user_cerberus_path",
    password="password_cerberus_path",
)
def download(
    repo: str,
    files: Iterable[str] = ("**",),
    directory: str = "",
    branch: str = "",
    user: str = "",
    password: str = "",
    cerberus_url: str = "",
    user_cerberus_path: str = "",
    password_cerberus_path: str = "",
) -> List[str]:
    """
    Download files from a git repository and return a list of the files
    downloaded.

    Parameters:

    - repo (str): A git URL, as you would pass to `git clone`
    - files ([str]): One or more
      [glob patterns](https://docs.python.org/3/library/glob.html)
      or relative file paths
    - directory (str): The target directory (if not provided, the current
      directory is used)
    - branch (str): A branch from which to retrieve (if not provided,
      files will be retrieved from HEAD)
    - user (str) = ""
    - password (str) = ""
    - cerberus_url (str) = "": A Cerberus API URL
    - cerberus_user_path (str) = "":
      The path to a Cerberus secure drop box secret where the username can be
      found
    - cerberus_password_path (str) = "":
      The path to a Cerberus secure drop box secret where the password can be
      found
    """
    if isinstance(files, str):
        files = (files,)
    if not directory:
        directory = os.path.curdir
    if user or password:
        repo = update_url_user_password(repo, user, password)
    directory = os.path.abspath(directory)
    # Shallow clone into a temp directory
    temp_directory: str = mkdtemp(prefix="git_download_")
    check_call(
        (
            ("git", "clone", "-q", "--depth", "1", "--single-branch")
            + (("-b", branch) if branch else ())
            + (repo, temp_directory)
        )
    )
    # Remove the git directory, so those files aren't accidentally matched
    rmtree(os.path.join(temp_directory, ".git"), ignore_errors=True)
    current_directory: str = os.path.abspath(os.path.curdir)
    path: str
    try:
        os.chdir(temp_directory)
        matched_files: Tuple[str, ...] = tuple(
            filter(
                os.path.isfile,
                map(
                    lambda path: os.path.join(
                        temp_directory, path  # type: ignore
                    ),
                    chain(*map(_iglob_recursive, files)),  # type: ignore
                ),
            )
        )
    finally:
        os.chdir(current_directory)
    downloaded_paths: List[str] = []
    new_path: str
    for path in matched_files:
        relative_path: str = os.path.relpath(path, temp_directory)
        new_path = os.path.join(directory, relative_path)
        if os.path.sep in relative_path:
            os.makedirs(os.path.dirname(new_path), exist_ok=True)
        print(new_path)
        os.rename(path, new_path)
        downloaded_paths.append(new_path)
    rmtree(temp_directory, ignore_errors=True)
    return downloaded_paths


def main() -> None:
    parser: argparse.ArgumentParser = argparse.ArgumentParser(
        prog="daves-dev-tools git download",
        description=(
            "Download files from a git repository matching one or more "
            "specified file names or glob patterns"
        ),
    )
    parser.add_argument(
        "-b",
        "--branch",
        default="",
        type=str,
        help=("Retrieve files from BRANCH instead of the remote's HEAD"),
    )
    parser.add_argument(
        "-d",
        "--directory",
        default="",
        type=str,
        help=(
            "The directory under which to save matched files. "
            "If not provided, files will be saved under the current "
            "directory."
        ),
    )
    parser.add_argument(
        "-u",
        "--user",
        default="",
        type=str,
        help="A username for accessing the repository",
    )
    parser.add_argument(
        "-p",
        "--password",
        default="",
        type=str,
        help="A password for accessing the repository",
    )
    parser.add_argument(
        "-cu",
        "--cerberus-url",
        default="",
        type=str,
        help=(
            "The base URL of a Cerberus REST API. See: https://swoo.sh/3DBW2Vb"
        ),
    )
    parser.add_argument(
        "-ucp",
        "--user-cerberus-path",
        default="",
        type=str,
        help=(
            "A Cerberus secure data path (including /key) wherein a "
            "username with which to authenticate can be found."
        ),
    )
    parser.add_argument(
        "-pcp",
        "--password-cerberus-path",
        default="",
        type=str,
        help=(
            "A Cerberus secure data path (including /key) wherein a "
            "password with which to authenticate can be found."
        ),
    )
    parser.add_argument("repo", type=str, help="Reference repository")
    parser.add_argument(
        "file",
        nargs="*",
        type=str,
        help=(
            "One or more `glob` pattern(s) indicating a specific file or "
            "files to include. If not provided, all files "
            "in the repository will be included."
        ),
    )
    arguments: argparse.Namespace = parser.parse_args()
    download(
        arguments.repo,
        files=arguments.file or ("**",),
        directory=arguments.directory,
        branch=arguments.branch,
        cerberus_url=arguments.cerberus_url,
        user=arguments.user,
        password=arguments.password,
        user_cerberus_path=arguments.user_cerberus_path,
        password_cerberus_path=arguments.password_cerberus_path,
    )


if __name__ == "__main__":
    main()



import sys
from importlib import import_module
from types import ModuleType

from . import __name__ as _module_name


def _print_help() -> None:
    print(
        "Usage:\n"
        "  daves-dev-tools requirements <command> [options]\n\n"
        "Commands:\n"
        "  update                      Update requirement versions in the "
        "specified files to\n"
        "                              align with currently installed "
        "versions of each distribution.\n"
        "  freeze                      Print dependencies inferred from an "
        "installed distribution\n"
        "                              or project, in a similar format to the "
        "output of `pip freeze`."
    )


def _get_command() -> str:
    command: str = ""
    if len(sys.argv) > 1:
        command = sys.argv.pop(1).lower().replace("-", "_")
    return command


def main() -> None:
    """
    Run a sub-module `main` function.
    """
    command = _get_command()
    module: ModuleType
    try:
        try:
            module = import_module(f"{_module_name}.{command}.__main__")
        except ImportError:
            module = import_module(f"{_module_name}.{command}")
        module.main()  # type: ignore
    except Exception:
        _print_help()


if __name__ == "__main__":
    main()



import argparse
import re
import sys
import os
from itertools import chain
from pkg_resources import Distribution
from pipes import quote
from typing import Iterable, Set, List, Tuple, Pattern, Sequence
from .requirements.utilities import (
    get_distribution,
    get_installed_distributions,
    get_requirements_required_distribution_names,
    normalize_name,
    get_setup_distribution_name,
    is_installed,
)
from .utilities import iter_parse_delimited_values, iter_sys_argv_pop, run


_SETUP_NAMES: Set[str] = {"setup.cfg", "setup.py"}
EXCLUDE_DIRECTORY_REGULAR_EXPRESSIONS: Tuple[str, ...] = (
    r"^[.~].*$",
    r"^venv$",
    r"^site-packages$",
)


def _get_requirement_string(
    name: str, directory: str, include_extras: bool
) -> str:
    requirement_string: str = os.path.abspath(directory)
    if is_installed(name) and include_extras:
        distribution: Distribution = get_distribution(name)
        if distribution.extras:
            requirement_string = (
                f"{requirement_string}" f"[{','.join(distribution.extras)}]"
            )
    return requirement_string


def _iter_find_distributions(
    distribution_names: Set[str],
    directories: Iterable[str] = ("../",),
    exclude_directory_regular_expressions: Iterable[
        str
    ] = EXCLUDE_DIRECTORY_REGULAR_EXPRESSIONS,
    include_extras: bool = False,
) -> Iterable[str]:
    if isinstance(directories, str):
        directories = (directories,)
    directories = map(os.path.abspath, directories)
    exclude_directory_patterns: Tuple[Pattern, ...] = tuple(
        map(re.compile, exclude_directory_regular_expressions)
    )

    def include_directory(directory: str) -> bool:
        directory_basename: str = os.path.basename(directory)
        for exclude_directory_pattern in exclude_directory_patterns:
            if exclude_directory_pattern.match(directory_basename):
                return False
        return True

    def iter_find_directory_distributions(directory: str) -> Iterable[str]:
        sub_directories: List[str]
        files: List[str]
        sub_directories, files = next(iter(os.walk(directory)))[1:3]

        def get_subdirectory_path(subdirectory: str) -> str:
            return os.path.join(directory, subdirectory)

        sub_directories = list(map(get_subdirectory_path, sub_directories))
        # Check to see if this is a project directory
        if any(map(_SETUP_NAMES.__contains__, map(str.lower, files))):
            name: str = get_setup_distribution_name(directory)
            if name in distribution_names:
                return (
                    _get_requirement_string(name, directory, include_extras),
                )
            else:
                return ()
        else:
            return chain(
                *map(
                    iter_find_directory_distributions,
                    filter(include_directory, sub_directories),
                )
            )

    return chain(*map(iter_find_directory_distributions, directories))


def find_and_install_distributions(
    distribution_names: Set[str],
    directories: Iterable[str] = ("../",),
    exclude_directory_regular_expressions: Iterable[
        str
    ] = EXCLUDE_DIRECTORY_REGULAR_EXPRESSIONS,
    dry_run: bool = False,
    include_extras: bool = False,
    pip_install_arguments: Sequence[str] = (),
) -> None:
    requirements: Tuple[str, ...] = tuple(
        map(
            quote,
            _iter_find_distributions(
                distribution_names=distribution_names,
                directories=directories,
                exclude_directory_regular_expressions=(
                    exclude_directory_regular_expressions
                ),
                include_extras=include_extras,
            ),
        )
    )
    if requirements:
        command: Tuple[str, ...] = (
            sys.executable,
            "-m",
            "pip",
            "install",
        ) + tuple(chain(*zip(("-e",) * len(requirements), requirements)))
        if pip_install_arguments:
            if isinstance(pip_install_arguments, str):
                pip_install_arguments = (pip_install_arguments,)
            else:
                pip_install_arguments = tuple(pip_install_arguments)
            if (
                "--upgrade-strategy" in pip_install_arguments
                and "-U" not in pip_install_arguments
                and "--upgrade" not in pip_install_arguments
            ):
                pip_install_arguments += ("-U",)
            command += pip_install_arguments
        if dry_run:
            print(" ".join(map(quote, command)))
        else:
            run(command)


def install_editable(
    requirements: Iterable[str] = (),
    directories: Iterable[str] = ("../"),
    exclude: Iterable[str] = (),
    exclude_directory_regular_expressions: Iterable[
        str
    ] = EXCLUDE_DIRECTORY_REGULAR_EXPRESSIONS,
    dry_run: bool = False,
    include_extras: bool = False,
    pip_install_arguments: Sequence[str] = (),
) -> None:
    """
    Install, in editable/develop mode, all distributions, except for those
    specified in `exclude`, which are required for the specified
    `requirements`.

    Parameters:
    - requirements ([str]) = ():
      One or more requirement specifiers or configuration file paths to which
      installation should be limited
    - directories ([str]) = ("../",): The directories in which to search
      for distributions to install. By default, the parent of the currently
      directory is used.
    - exclude ([str]): One or more distributions to pass over when searching
      for distributable projects
    - exclude_directory_regular_expressions ([str])
    - dry_run (bool)
    - include_extras (bool)
    - pip_install_arguments ([str]): Additional arguments to pass on to
      `pip install`
    """
    required_distribution_names: Set[str] = (
        get_requirements_required_distribution_names(requirements)
        if requirements
        else set(get_installed_distributions().keys())
    )
    find_and_install_distributions(
        distribution_names=(
            required_distribution_names - set(map(normalize_name, exclude))
        ),
        directories=directories,
        exclude_directory_regular_expressions=(
            exclude_directory_regular_expressions
        ),
        dry_run=dry_run,
        include_extras=include_extras,
        pip_install_arguments=pip_install_arguments,
    )


def main() -> None:
    parser: argparse.ArgumentParser = argparse.ArgumentParser(
        prog="daves-dev-tools install-editable",
        description=(
            "This command will attempt to find and install, in "
            "develop (editable) mode, all packages which are "
            "installed in the current python environment. If one or "
            "more `requirement` file paths or specifiers are provided, "
            "installation will be limited to the dependencies identified "
            "(recursively) by these requirements. Exclusions can be specified "
            "using the `-e` parameter. Directories can be excluded by "
            "passing regular expressions to the `-edre` parameter. "
            "Any arguments passed not matching those specified below "
            "will be passed on to `pip install` (see `pip install -h` for "
            "additionally available arguments)."
        ),
    )
    parser.add_argument(
        "-r",
        "--requirement",
        action="append",
        type=str,
        default=[],
        help=(
            "One or more requirement specifiers, requirement file paths, "
            "or configuration file paths "
            "(setup.cfg, setup.py, pyproject.toml, tox.ini, etc.). "
            "If provided, only dependencies of these requirements will be "
            "installed."
        ),
    )
    parser.add_argument(
        "-d",
        "--directory",
        default=["../"],
        type=str,
        action="append",
        help=(
            "A directory in which to search for requirements. "
            "By default, the directory above the current directory is "
            "searched. This argument may be passed more than once to include "
            "multiple locations."
        ),
    )
    parser.add_argument(
        "-e",
        "--exclude",
        default=[],
        type=str,
        action="append",
        help="A comma-separated list of distribution names to exclude",
    )

    parser.add_argument(
        "-edre",
        "--exclude-directory-regular-expression",
        default=list(EXCLUDE_DIRECTORY_REGULAR_EXPRESSIONS),
        type=str,
        action="append",
        help=(
            "Directories matching this regular expression will be excluded "
            "when searching for setup locations This argument may be passed "
            "more than once to exclude directories matching more than one "
            "regular expression. The default for this argument is "
            "equivalent to `-edre {}`".format(
                " -edre ".join(
                    map(quote, EXCLUDE_DIRECTORY_REGULAR_EXPRESSIONS)
                )
            )
        ),
    )
    parser.add_argument(
        "-dr",
        "--dry-run",
        default=False,
        action="store_const",
        const=True,
        help="Print, but do not execute, all `pip install` commands",
    )
    parser.add_argument(
        "-ie",
        "--include-extras",
        default=False,
        action="store_const",
        const=True,
        help="Install all extras for all discovered distributions",
    )
    # For backwards compatibility, we accept requirements as either
    # positional or keyword arguments
    positional_arguments: List[str] = list(iter_sys_argv_pop())
    namespace: argparse.Namespace
    unknown_arguments: List[str]
    namespace, unknown_arguments = parser.parse_known_args()
    install_editable(
        requirements=namespace.requirement + positional_arguments,
        directories=namespace.directory,
        exclude_directory_regular_expressions=(
            namespace.exclude_directory_regular_expression
        ),
        exclude=iter_parse_delimited_values(namespace.exclude),
        dry_run=namespace.dry_run,
        include_extras=namespace.include_extras,
        pip_install_arguments=unknown_arguments,
    )


if __name__ == "__main__":
    main()



import functools
import logging
import sys
from inspect import Parameter, Signature, signature
from warnings import warn
from collections import OrderedDict
from traceback import format_exception
from typing import (
    Any,
    Callable,
    Dict,
    List,
    Optional,
    Tuple,
    Iterable,
    Mapping,
    Union,
)
from .utilities import sys_argv_pop, iter_sys_argv_pop
from .errors import append_exception_text

__all__: List[str] = [
    "get_cerberus_secrets",
    "get_cerberus_secret",
    "apply_sys_argv_cerberus_arguments",
]
lru_cache: Callable[..., Any] = functools.lru_cache


@lru_cache()
def get_cerberus_secrets(
    cerberus_url: str,
    path: str,
) -> Dict[str, str]:
    """
    This function attempts to access Cerberus secrets at the given `path`
    with each AWS profile until successful, or until having run out of
    profiles, and returns the secrets if successful (or raises an error if
    not).

    Parameters:

    - **cerberus_url** (str): The Cerberus API endpoint
    - **path** (str): The Cerberus path containing the desired secret(s)
    """
    try:
        import boto3  # type: ignore
        from botocore.exceptions import (  # type: ignore
            ClientError,
            NoCredentialsError,
        )
        from cerberus import CerberusClientException  # type: ignore
        from cerberus.client import CerberusClient  # type: ignore
    except ImportError as error:
        append_exception_text(
            error,
            (
                'Please install *daves-dev-tools* with the "cerberus" extra: '
                "`pip3 install 'daves-dev-tools[cerberus]'`"
            ),
        )
        raise error
    secrets: Optional[Dict[str, str]] = None
    errors: Dict[str, str] = OrderedDict()
    for profile_name in boto3.Session().available_profiles + [None]:
        arn: str = ""
        try:
            session: boto3.Session = boto3.Session(profile_name=profile_name)
            arn = session.client("sts").get_caller_identity().get("Arn")
            secrets = CerberusClient(
                cerberus_url, aws_session=session
            ).get_secrets_data(path)
            break
        except (CerberusClientException, ClientError, NoCredentialsError):
            errors[arn or profile_name] = "".join(
                format_exception(*sys.exc_info())
            )
    if secrets is None:
        error_text: str = "\n".join(
            f'{key or "[default]"}:\n{value}\n'
            for key, value in errors.items()
        )
        warn(error_text)
        logging.warning(error_text)
        raise PermissionError(
            "No AWS profile was found with access to the requested secrets"
        )
    return secrets


def get_cerberus_secret(cerberus_url: str, path: str) -> Tuple[str, str]:
    """
    This is a convenience function for retrieving individual values from a
    Cerberus secret, and which return both the key (the last part of the
    path), as well as the retrieved secret value, as a `tuple`. This is
    useful under a common scenario where a username is a key in a secrets
    dictionary, and the client application needs both the username as password.

    Parameters:

    - **cerberus_url** (str): The Cerberus API endpoint
    - **path** (str): The Cerberus path containing the desired secret,
      *including* a dictionary key. For example: "path/to/secrets/key".
    """
    assert cerberus_url
    parts: List[str] = path.strip("/").split("/")
    # Ensure the path includes the dictionary key
    assert len(parts) == 4
    key: str = parts.pop()
    # Retrieve the secrets dictionary
    secrets: Dict[str, str] = get_cerberus_secrets(
        cerberus_url, "/".join(parts)
    )
    # Return the key and the value stored at the secret's "key" index
    return key, secrets[key]


def apply_sys_argv_cerberus_arguments(
    url_parameter_name: Iterable[str],
    parameter_map: Union[
        Mapping[str, Iterable[str]],
        Iterable[Tuple[str, Iterable[str]]],
    ],
    argv: Optional[List[str]] = None,
) -> None:
    """
    This function modifies `sys.argv` by performing lookups against a specified
    Cerberus API and inserting the retrieved values into mapped keyword
    argument values.

    Parameters:

    - url_parameter_name (str): The base URL of the Cerberus API.
    - parameter_map ({str: [str]}): Maps final keyword argument names to
      keyword argument names wherein to find Cerberus paths to lookup
      values to pass to the final keyword arguments. Cerberus path keyword
      arguments are removed from `sys.argv` and final keyword arguments are
      inserted into `sys.argv`.
    - argv ([str]) = sys.argv: If provided, this list will be modified instead
      of `sys.argv`.
    """
    if argv is None:
        argv = sys.argv
    parameter_map_items: Iterable[Tuple[str, Iterable[str]]]
    if isinstance(parameter_map, Mapping):
        parameter_map_items = parameter_map.items()
    else:
        parameter_map_items = parameter_map
    cerberus_url: str = sys_argv_pop(  # type: ignore
        keys=url_parameter_name, flag=False
    )
    if cerberus_url:
        key: str
        cerberus_path_keys: Iterable[str]
        for key, cerberus_path_keys in parameter_map_items:
            value: str
            for value in iter_sys_argv_pop(  # type: ignore
                keys=cerberus_path_keys, argv=argv, flag=False
            ):
                argv += [key, get_cerberus_secret(cerberus_url, value)[-1]]


def _merge_function_signature_args_kwargs(
    function_signature: Signature, args: Iterable[Any], kwargs: Dict[str, Any]
) -> None:
    """
    This function merges positional/keyword arguments for a function
    into the keyword argument dictionary
    """
    value: Any
    parameter: Parameter
    if args:
        for parameter, value in zip(
            function_signature.parameters.values(), args
        ):
            assert parameter.kind == Parameter.POSITIONAL_OR_KEYWORD
            kwargs[parameter.name] = value


def _remove_function_signature_inapplicable_kwargs(
    function_signature: Signature, kwargs: Dict[str, Any]
) -> None:
    def get_parameter_name(parameter_: Parameter) -> str:
        return parameter_.name

    key: str
    for key in set(kwargs.keys()) - set(
        map(
            get_parameter_name,
            function_signature.parameters.values(),
        )
    ):
        del kwargs[key]


def apply_cerberus_path_arguments(
    cerberus_url_parameter_name: str = "cerberus_url",
    **cerberus_path_parameter_names: str,
) -> Callable[..., Callable[..., Any]]:
    """
    This decorator maps parameter names. Each key represents the
    name of a parameter in the decorated function which accepts an explicit
    input, and the corresponding mapped value is the name of a second parameter
    which accepts a cerberus path from where a value for the first parameter
    can be retrieved when not explicitly provided.

    Parameters:
    - cerberus_url_parameter_name (str) = "cerberus_url":
      The name of the Cerberus API URL parameter
    - ** (str): All additional keyword map a
      *parameter name* -> *cerberus URL parameter name*
    """

    def decorating_function(
        function: Callable[..., Any]
    ) -> Callable[..., Any]:
        function_signature: Signature = signature(function)

        @functools.wraps(function)
        def wrapper(*args: Any, **kwargs: Any) -> Any:
            """
            This function wraps the original and performs lookups for
            any parameters for which an argument is not passed
            """
            # First we consolidate the keyword arguments with any arguments
            # which are passed to parameters which can be either positional
            # *or* keyword arguments, and were passed as positional arguments
            _merge_function_signature_args_kwargs(
                function_signature, args, kwargs
            )
            cerberus_url: str = ""
            if cerberus_url_parameter_name in kwargs:
                cerberus_url = kwargs[cerberus_url_parameter_name]
            elif cerberus_url_parameter_name in function_signature.parameters:
                cerberus_url = (
                    function_signature.parameters[
                        cerberus_url_parameter_name
                    ].default
                    or ""
                )
            # For any arguments where we have a cerberus path and do not have
            # an explicitly passed value, perform a lookup in cerberus
            key: str
            for key in set(cerberus_path_parameter_names.keys()) - set(
                kwargs.keys()
            ):
                cerberus_path_key: str = cerberus_path_parameter_names[key]
                if (cerberus_path_key in kwargs) and kwargs[cerberus_path_key]:
                    kwargs[key] = get_cerberus_secret(
                        cerberus_url,
                        kwargs[cerberus_path_key],
                    )[1]
                elif cerberus_path_key in function_signature.parameters:
                    default: Optional[str] = function_signature.parameters[
                        cerberus_path_key
                    ].default
                    if default:
                        kwargs[key] = get_cerberus_secret(
                            cerberus_url,
                            default,
                        )[1]
            # Remove arguments which do not correspond to
            # any of the function's parameter names
            _remove_function_signature_inapplicable_kwargs(
                function_signature, kwargs
            )
            # Execute the wrapped function
            return function(**kwargs)

        return wrapper

    return decorating_function



"""
This module cleans up files which are ignored by git
"""
import argparse
import os
import shutil
from subprocess import check_output, check_call
from collections import deque
from glob import iglob
from itertools import chain
from typing import (
    Dict,
    Tuple,
    Iterable,
    Sequence,
    Set,
    FrozenSet,
    List,
)

ROOT_DIRECTORY: str = "."
DEFAULT_EXCLUDE: Tuple[str, ...] = (
    ".idea",  # Jetbrains' IDE project settings (Pycharm, Intellij IDEA)
    ".vscode",  # Microsoft Visual Studio Code project settings
    ".git",  # Git history
    "venv",  # Commonly used location for virtual environments
)


def _get_directory_globs_files(
    directory: str, patterns: Iterable[str], *, recursive: bool = False
) -> FrozenSet[str]:
    """
    Return a `frozenset` of file names matching the glob `pattern` within
    `directory`, or in a directory matching the glob pattern.

    Parameters:

    - directory (str)
    - patterns ([str])
    - recursive (bool) = False
    """
    if isinstance(patterns, str):
        patterns = (patterns,)
    file_relative_paths: Set[str]
    relative_paths: FrozenSet[str]
    current_directory: str = os.path.curdir
    os.chdir(directory)
    try:
        relative_paths = frozenset()

        def add_glob(pattern: str) -> None:
            nonlocal relative_paths
            relative_paths |= frozenset(iglob(pattern, recursive=recursive))

        deque(map(add_glob, patterns), maxlen=0)
        file_relative_paths = set()
        relative_path: str
        for relative_path in relative_paths:
            if os.path.isdir(relative_path):
                sub_directory: str
                _: Iterable[str]
                files: Iterable[str]
                for sub_directory, _, files in os.walk(relative_path):
                    sub_directory = sub_directory.replace("\\", "/")
                    file: str
                    for file in files:
                        file_relative_paths.add(f"{sub_directory}/{file}")
            else:
                file_relative_paths.add(relative_path.replace("\\", "/"))
    finally:
        os.chdir(current_directory)
    return frozenset(file_relative_paths)


def _get_directory_globs(
    directory: str, patterns: Iterable[str], *, recursive: bool = False
) -> FrozenSet[str]:
    """
    Return a `frozenset` of directory paths matching the glob `pattern` within
    `directory`, or in a directory matching the glob pattern.

    Parameters:

    - directory (str)
    - patterns ([str])
    - recursive (bool) = False
    """
    if isinstance(patterns, str):
        patterns = (patterns,)
    relative_paths: FrozenSet[str]
    current_directory: str = os.path.curdir
    os.chdir(directory)
    try:
        relative_paths = frozenset()

        def add_glob(pattern: str) -> None:
            nonlocal relative_paths
            relative_paths |= frozenset(iglob(pattern, recursive=recursive))

        deque(map(add_glob, patterns), maxlen=0)
    finally:
        os.chdir(current_directory)
    return relative_paths


def get_ignored_files(
    directory: str = ".",
    exclude: FrozenSet[str] = frozenset(),
) -> Set[str]:
    """
    Get a `set` containing the relative paths of all ignored files
    in `directory` excluding those matching any of the glob patterns
    in `exclude`.

    Parameters:

    - directory (str): The root project directory.
    - exclude ({str}) = frozenset(): A `frozenset` of glob patterns for
      files and sub-directories to exclude.
    """
    directory = os.path.abspath(directory)
    check_call(("git", "init", directory))
    check_call(("git", "add", directory))
    return set(
        check_output(
            ("git", "ls-files", "-o", directory),
            encoding="utf-8",
            universal_newlines=True,
        )
        .strip()
        .split("\n")
    ) - _get_directory_globs_files(directory, exclude, recursive=True)


def _is_sub_directory_excluded(
    sub_directory: str, directory: str, exclude: FrozenSet[str]
) -> bool:
    relative_sub_directory: str = os.path.relpath(
        os.path.abspath(sub_directory), directory
    )
    if relative_sub_directory.startswith("./"):
        relative_sub_directory = relative_sub_directory[2:]
    relative_sub_directory_list: List[str] = relative_sub_directory.split("/")
    index: int
    for index in range(1, len(relative_sub_directory_list)):
        ancestor: str = "/".join(relative_sub_directory_list[:index])
        if ancestor in exclude:
            return True
    return False


def delete_empty_directories(
    directory: str = ".",
    exclude: FrozenSet[str] = frozenset(),
    dry_run: bool = False,
    _recurrence: bool = False,
) -> int:
    """
    Deletes empty directories under the current directory.

    Parameters:

    - exclude ({str}) = frozenset():
      A set of top-level directory names to exclude
    """
    number_of_deleted_directories: int = 0
    sub_directory: str
    sub_directories: Sequence[str]
    files: Sequence[str]
    exclude = _get_directory_globs(directory, exclude)
    for sub_directory, sub_directories, files in os.walk(
        directory, topdown=False
    ):
        if not (
            _is_sub_directory_excluded(sub_directory, directory, exclude)
            or any(
                filter(
                    lambda name: name != ".DS_Store",
                    chain(sub_directories, files),
                )
            )
        ):
            if dry_run:
                print(f"rm -R {sub_directory}")
            else:
                shutil.rmtree(sub_directory)
            number_of_deleted_directories += 1
    if number_of_deleted_directories and not dry_run:
        number_of_deleted_directories += delete_empty_directories(
            directory,
            exclude=exclude,
            dry_run=dry_run,
            _recurrence=True,
        )
    if (not _recurrence) and number_of_deleted_directories:
        if not dry_run:
            print(f"Deleted {number_of_deleted_directories} empty directories")
    return number_of_deleted_directories


def delete_ignored(
    directory: str = ".",
    exclude: FrozenSet[str] = frozenset(),
    dry_run: bool = False,
) -> None:
    """
    Delete files which are ignored by Git.

    Parameters:

    - root_directory (str): The root project directory.
    - exclude ({{str}}) = {EXCLUDE_DIRECTORIES}: A `frozenset` of
      directories to leave untouched.
    """
    sub_directory_name: str
    directories_files: Dict[str, Set[str]] = {}
    paths: Set[str]
    path: str
    for path in get_ignored_files(directory=directory, exclude=exclude):
        sub_directory_name = ""
        if "/" in path:
            sub_directory_name = path.split("/")[0]
        if sub_directory_name not in directories_files:
            directories_files[sub_directory_name] = set()
        directories_files[sub_directory_name].add(
            os.path.join(directory, path)
        )
    for sub_directory_name, paths in directories_files.items():
        number_of_files: int = len(paths)
        if not dry_run:
            print(
                f"Deleting {number_of_files} ignored "
                f'file{"s" if number_of_files > 1 else ""} in '
                f"{sub_directory_name}"
            )
        for path in paths:
            if dry_run:
                print(f"rm {path}")
            else:
                os.remove(path)


delete_ignored.__doc__ = delete_ignored.__doc__.format(  # type: ignore
    EXCLUDE_DIRECTORIES=repr(frozenset(DEFAULT_EXCLUDE))
)


def clean(
    directory: str = ".",
    exclude: FrozenSet[str] = frozenset(DEFAULT_EXCLUDE),
    dry_run: bool = False,
) -> None:
    """
    Cleanup (delete) files which are ignored by Git and subsequently delete all
    empty directories.

    Parameters:

    - root_directory (str) = ".": The project's root directory.
    - exclude ({{str}}) = {EXCLUDE_DIRECTORIES}: A `frozenset` of
      directories to leave untouched.
    """
    delete_ignored(directory, exclude=exclude, dry_run=dry_run)
    delete_empty_directories(directory, exclude=exclude, dry_run=dry_run)


clean.__doc__ = clean.__doc__.format(  # type: ignore
    EXCLUDE_DIRECTORIES=repr(frozenset(DEFAULT_EXCLUDE))
)


def main() -> None:
    parser: argparse.ArgumentParser = argparse.ArgumentParser(
        prog="daves-dev-tool clean",
        description=(
            "This command removes all files from your project directory which "
            "are ignored by git, unless matching one of the EXCLUDE glob "
            "patterns."
        ),
    )
    exclude: str
    default_exclude_parameters: str = " ".join(
        f"-e {exclude}" for exclude in DEFAULT_EXCLUDE
    )
    parser.add_argument(
        "-e",
        "--exclude",
        type=str,
        action="append",
        help=(
            "One or more glob patterns indicating files/directories to "
            "exclude from cleanup. The default values are: `"
            f"{default_exclude_parameters}`."
        ),
    )
    parser.add_argument(
        "-dr",
        "--dry-run",
        default=False,
        action="store_const",
        const=True,
        help=(
            "Instead of executing the cleanup, just print the shell "
            "commands (a list of `rm FILE` commands)."
        ),
    )
    parser.add_argument(
        "directory",
        type=str,
        default=".",
        nargs="?",
        help="The root directory for the project.",
    )
    arguments: argparse.Namespace = parser.parse_args()
    clean(
        directory=arguments.directory,
        exclude=frozenset(
            arguments.exclude if arguments.exclude else DEFAULT_EXCLUDE
        ),
        dry_run=arguments.dry_run,
    )


if __name__ == "__main__":
    main()



import functools
import os
import re
import sys
from distutils.core import run_setup
from time import time
from typing import (
    Any,
    Callable,
    FrozenSet,
    Iterable,
    List,
    Optional,
    Tuple,
)
from subprocess import check_output
from .utilities import sys_argv_get, sys_argv_pop, run_module_as_main

from .cerberus import apply_sys_argv_cerberus_arguments

lru_cache: Callable[..., Any] = functools.lru_cache


def _list_dist(
    directory: str, modified_at_or_after: float = 0.0
) -> FrozenSet[str]:
    dist_root: str = os.path.join(directory, "dist")
    dist_file: str
    dist_sub_directories: List[str]
    dist_files: Iterable[str]
    try:
        dist_root, dist_sub_directories, dist_files = next(
            iter(os.walk(dist_root))
        )
    except StopIteration:
        raise FileNotFoundError(
            f"No distributions could be found in {dist_root}"
        )
    dist_files = (
        os.path.join(dist_root, dist_file) for dist_file in dist_files
    )
    if modified_at_or_after:
        dist_files = filter(
            lambda dist_file: (  # noqa
                os.path.getmtime(dist_file) >= modified_at_or_after
            ),
            dist_files,
        )
    try:
        return frozenset(dist_files)
    except (NotADirectoryError, FileNotFoundError):
        return frozenset()


def _setup(directory: str) -> FrozenSet[str]:
    start_time: float = time()
    current_directory: str = os.path.abspath(os.path.curdir)
    os.chdir(directory)
    try:
        abs_setup: str = os.path.join(directory, "setup.py")
        setup_args: List[str] = ["sdist", "bdist_wheel"]
        print(f'{sys.executable} {abs_setup} {" ".join(setup_args)}')
        run_setup(abs_setup, setup_args)
    finally:
        os.chdir(current_directory)
    return _list_dist(directory, modified_at_or_after=start_time)


def _get_help() -> bool:
    """
    If `-h` or `--help` keyword arguments are provided,
    retrieve the repository credentials and store them in the "TWINE_USERNAME"
    and "TWINE_PASSWORD" environment variables.
    """
    if set(sys.argv) & {"-h", "--help", "-H", "--HELP"}:
        help_: str = check_output(
            (sys.executable, "-m", "twine", "upload", "-h"),
            encoding="utf-8",
            universal_newlines=True,
        ).strip()
        help_ = re.sub(
            r"\btwine upload\b( \[-h\])?",
            (
                "daves-dev-tools distribute\\1 "
                "[-cu CERBERUS_URL]\n"
                "                    [-cup CERBERUS_USERNAME_PATH]\n"
                "                    [-cpp CERBERUS_PASSWORD_PATH]\n"
                "                   "
            ),
            help_,
        )
        help_ = re.sub(
            (
                r"(\n\s*)dist \[dist \.\.\.\](?:.|\n)+"
                r"(\npositional arguments:\s*\n\s*)(?:.|\n)+"
                r"(\noptional arguments:\s*\n)"
            ),
            (
                r"\1[directory]"
                r"\n\2directory             "
                "The root directory path for the project."
                r"\n\3"
            ),
            help_,
        )
        help_ = (
            f"{help_.rstrip()}\n"
            "  -cu CERBERUS_URL, --cerberus-url CERBERUS_URL\n"
            "                        The base URL of a Cerberus REST API.\n"
            "                        See: https://swoo.sh/3DBW2Vb\n"
            "  -cup CERBERUS_USERNAME_PATH, --cerberus-username-path "
            "CERBERUS_USERNAME_PATH\n"
            "                        A Cerberus secure data path (including "
            "/key) wherein a\n"
            "                        username with which to authenticate can "
            "be found.\n"
            "                        See: https://swoo.sh/3DBW2Vb\n"
            "  -cpp CERBERUS_PASSWORD_PATH, --cerberus-password-path "
            "CERBERUS_PASSWORD_PATH\n"
            "                        A Cerberus secure data path (including "
            "/key) wherein a\n"
            "                        password with which to authenticate can "
            "be found.\n"
            "                        If no USERNAME or CERBERUS_USERNAME_PATH "
            "is provided,\n"
            "                        the last part of this path \n"
            "                        (the secure data path entry key) is "
            "inferred as your\n"
            "                        username. See: https://swoo.sh/3DBW2Vb\n"
        )
        print(help_)
        return True
    return False


def _get_credentials_from_cerberus() -> None:
    """
    If `--cerberus-url` and `--cerberus-path` keyword arguments are provided,
    retrieve the repository credentials and apply them to their corresponding
    static arguments.
    """
    cerberus_password_path_keywords: Tuple[str, str, str, str] = (
        "-cpp",
        "--cerberus-password-path",
        # For backwards compatibility
        "-cp",
        "--cerberus-path",
    )
    username: Optional[str] = sys_argv_get(  # type: ignore
        ("-u", "--username")
    )
    cerberus_password_path: Optional[str] = sys_argv_get(  # type: ignore
        cerberus_password_path_keywords
    )
    if cerberus_password_path:
        cerberus_password_path = cerberus_password_path.strip("/ ")
        cerberus_password_path_list: List[str] = cerberus_password_path.split(
            "/"
        )
        path_length: int = len(cerberus_password_path_list)
        if path_length == 3 and username:
            # Append the SDB key
            sys_argv_pop(cerberus_password_path_keywords, flag=False)
            sys.argv += [
                cerberus_password_path_keywords[0],
                f"{cerberus_password_path}/{username}",
            ]
        elif path_length == 4:
            if not username:
                # Infer the username to be the SDB key
                sys.argv += ["-u", cerberus_password_path_list[-1]]
        else:
            raise ValueError(
                "The value for -cpp or --cerberus-password-path must be "
                "formatted either as:\n"
                '- "namespace/safe-deposit-box/secret" '
                "(if a `--username` is provided) or\n"
                '- "namespace/safe-deposit-box/secret/key"\n'
                f"...not: {repr(cerberus_password_path)}"
            )
    apply_sys_argv_cerberus_arguments(
        ("-cu", "--cerberus-url"),
        {
            "-u": ("-cup", "--cerberus-username-path"),
            "-p": cerberus_password_path_keywords,
        },
    )


def _dist(
    directory: str, distributions: FrozenSet[str], echo: bool = True
) -> None:
    run_module_as_main(
        "twine",
        arguments=(["upload"] + sys.argv[1:] + list(sorted(distributions))),
        directory=directory,
        echo=False,
    )


def _cleanup(directory: str) -> None:
    current_directory: str = os.path.abspath(os.path.curdir)
    os.chdir(directory)
    try:
        run_setup(os.path.join(directory, "setup.py"), ["clean", "--all"])
    finally:
        os.chdir(current_directory)


def main() -> None:
    if not _get_help():
        _get_credentials_from_cerberus()
        directory: str = sys_argv_pop(depth=2, default=".")  # type: ignore
        directory = os.path.abspath(directory).rstrip("/")
        try:
            _dist(directory, _setup(directory))
        finally:
            _cleanup(directory)


if __name__ == "__main__":
    main()



import argparse
from collections import deque
import os
import re
from io import StringIO
from pathlib import Path
from typing import IO, Set
from typing import Tuple
from configparser import ConfigParser
from .requirements.utilities import (
    iter_distribution_location_file_paths,
)


def _get_project_and_setup_cfg_paths(path: str = ".") -> Tuple[str, str]:
    setup_cfg_path: str
    project_path: str
    if not os.path.isdir(path):
        assert os.path.basename(path).lower() == "setup.cfg"
        setup_cfg_path = path
        project_path = os.path.dirname(path)
    else:
        setup_cfg_path = os.path.join(path, "setup.cfg")
        project_path = path
    return project_path, setup_cfg_path


def _touch_packages_py_typed(project_path: str) -> None:
    py_typed_paths: Set[str] = set()
    setup_py_path: str = str(
        Path(project_path).absolute().joinpath("setup.py")
    )

    def touch_py_typed(path: str) -> None:
        if path.endswith(".py"):
            absolute_path: Path = Path(path).absolute()
            if str(absolute_path) != setup_py_path:
                py_typed_path: Path = absolute_path.parent.joinpath("py.typed")
                py_typed_path_str: str = str(py_typed_path)
                if py_typed_path_str not in py_typed_paths:
                    print(f"touch {py_typed_path}")
                    py_typed_path.touch()
                    py_typed_paths.add(py_typed_path_str)

    deque(
        map(
            touch_py_typed, iter_distribution_location_file_paths(project_path)
        ),
        maxlen=0,
    )


def _update_setup_cfg(
    setup_cfg_path: str,
) -> None:
    parser: ConfigParser = ConfigParser()
    if os.path.isfile(setup_cfg_path):
        parser.read(setup_cfg_path)
    if not parser.has_section("options"):
        parser.add_section("options")
    parser.set("options", "include_package_data", "True")
    if not parser.has_section("options.package_data"):
        parser.add_section("options.package_data")
    package_data: str = parser.get(
        "options.package_data", "*", fallback=""
    ).rstrip()
    if "py.typed" not in filter(  # type: ignore
        None,
        map(
            os.path.normpath,  # type: ignore
            map(str.strip, package_data.split("\n")),
        ),
    ):
        parser.set(
            "options.package_data",
            "*",
            "py.typed",
        )
    print(f"Writing {setup_cfg_path}")
    setup_cfg: str
    setup_cfg_io: IO[str]
    with StringIO() as setup_cfg_io:
        parser.write(setup_cfg_io)
        setup_cfg_io.seek(0)
        setup_cfg = re.sub(r"[ ]+(\n|$)", r"\1", setup_cfg_io.read()).strip()
        setup_cfg = f"{setup_cfg}\n"
    with open(setup_cfg_path, "w") as setup_cfg_io:
        setup_cfg_io.write(setup_cfg)


def make_typed(path: str = ".") -> None:
    """
    Create (if needed) **/py.typed files and alter the setup.cfg file such that
    a distribution's packages will be identified as being fully type-hinted
    """
    project_path: str
    setup_cfg_path: str
    project_path, setup_cfg_path = _get_project_and_setup_cfg_paths(path)
    # Create py.typed files
    _touch_packages_py_typed(project_path)
    # Parse and update setup.cfg
    _update_setup_cfg(setup_cfg_path)


def main() -> None:
    parser: argparse.ArgumentParser = argparse.ArgumentParser(
        prog="daves-dev-tools make-typed",
        description=(
            "Add **/py.typed files and alter the setup.cfg such that a "
            "distribution's packages will be identifiable as fully type-hinted"
        ),
    )
    parser.add_argument(
        "path",
        nargs="?",
        default=".",
        type=str,
        help=(
            "A project directory (where the setup.py and/or setup.cfg file "
            "are located)"
        ),
    )
    arguments: argparse.Namespace = parser.parse_args()
    make_typed(arguments.path)


if __name__ == "__main__":
    main()



import argparse
import os
from typing import Iterable
from subprocess import check_output, check_call
from ..requirements.utilities import get_setup_distribution_version


def tag_version(directory: str = os.path.curdir, message: str = "") -> None:
    """
    Tag your project with the package version number *if* no pre-existing
    tag with that version number exists.

    Parameters:

    - directory (str)
    - message (str)
    """
    version: str = get_setup_distribution_version(directory)
    if version:
        current_directory: str = os.path.abspath(os.path.curdir)
        os.chdir(directory)
        try:
            tags: Iterable[str] = map(
                str.strip,
                check_output(
                    ("git", "tag"), encoding="utf-8", universal_newlines=True
                )
                .strip()
                .split("\n"),
            )
            if version not in tags:
                check_call(
                    ("git", "tag", "-a", version, "-m", message or version)
                )
        finally:
            os.chdir(current_directory)


def main() -> None:
    parser: argparse.ArgumentParser = argparse.ArgumentParser(
        prog="daves-dev-tools git tag-version",
        description=(
            "Tag your repo with the python package version, if a tag "
            "for that version doesn't already exist."
        ),
    )
    parser.add_argument(
        "directory",
        nargs="?",
        default=os.path.curdir,
        type=str,
        help=(
            "Your project directory. If not provided, the current "
            "directory will be used."
        ),
    )
    parser.add_argument(
        "-m",
        "--message",
        default="",
        type=str,
        help=(
            "The tag message. If not provided, the new version number is "
            "used."
        ),
    )
    arguments: argparse.Namespace = parser.parse_args()
    tag_version(directory=arguments.directory, message=arguments.message)


if __name__ == "__main__":
    main()
